/*
 * (c) Copyright Robin Hill 2002
 *
 * For license terms, see the file COPYING in this directory.
 */

// For compilers that support precompilation, includes "wx/wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
	#pragma hdrstop
#endif

// For all others, include the necessary headers (this file is usually all you
// need because it includes almost all "standard" wxWindows headers
#ifndef WX_PRECOMP
	#include <wx/wx.h>
#endif

#ifdef __WINDOWS__
	#include <windows.h>
	#include <wincrypt.h>
	#include <io.h>
#else
	#include <unistd.h>
	#include <openssl/evp.h>
#endif

#include <wx/textfile.h>
#include <wx/socket.h>
#include <math.h>

#include "defines.h"
#include "frFFile.h"
#include "fr_backend.h"
#include "fr_backend_tiger.h"

#include <wx/listimpl.cpp>
WX_DEFINE_LIST(CRCList);
WX_DEFINE_LIST(ErrorList);

//
// init_crctab
//
// Initialises the CRC lookup table.  Precalculating this section
// provides a speedup at a very minor (1k) memory cost.
//
void FR_BackEnd::init_crctab(tcrc polynomial)
{
	wxLogTrace(wxT("frTRACE_functions"), wxT("init_crctab: Entry"));
	wxInt16 i, j;
	tcrc  crc;

	for(i = 0; i < 256; i++) {
		crc = i;

		for(j = 8; j > 0; j--) {
			if(crc & 1)
				crc = (crc >> 1) ^ polynomial;
			else
				crc >>= 1;
		}
		crctab[i] = crc;
	}
	wxLogTrace(wxT("frTRACE_functions"), wxT("init_crctab: Exit"));
}

//
// CreateFileHashED2k
//
// Receives a filename and buffer size as input.  Creates an eDonkey hash
// from the file contents (this is an MD4 based hash - each DEF_ED2KBLOCK
// from the file has an MD4 created for it.  All of these hashes are then
// appended and an MD4 hash is created for this.
//
void FR_BackEnd::CreateFileHashED2k(const wxString FileName, const wxUint32 bufsize)
{
	frFFile         *InputFile;
	wxUint32        read,
	                blockleft,
	                hashcnt,
	                cnt;
	unsigned char   *buffer,
	                *hashes,
	                *hash;
	wxString        message;
#ifdef __WINDOWS__
	unsigned long   hashlen;
	HCRYPTPROV      CryptoProvider;
	HCRYPTHASH      CryptoHash;
#endif
#ifdef __UNIX__
	wxUint32        hashlen;
	EVP_MD_CTX      md_temp;
#endif

	wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashED2k: Entry"));
	wxASSERT(bufsize > 0);
	wxASSERT(FileName.Len() > 0);
	SendStatus(wxString::Format(wxT(STAT_MD_CREATE), wxT("eDonkey"), FileName.c_str()));

// Open the input file
	wxLogDebug(wxString::Format(wxT("CreateFileHashED2k: opening file %s"), FileName.c_str()));
	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: allocating InputFile"));
	InputFile = new frFFile(FileName, wxT("rb"));
	if (! InputFile->IsOpened()) {
		SendError(wxString::Format(wxT(ERR_FILESTAT), FileName.c_str()));
		wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashED2k: Error"));
		return;
	}

// Calculate the number of DEF_ED2KBLOCK sized blocks in the file
	hashcnt = ((InputFile->Length() + DEF_ED2KBLOCK - 1) / DEF_ED2KBLOCK);

	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: allocating buffers"));
	buffer = new unsigned char[bufsize];
	hashes = new unsigned char[hashcnt * 16];
	hash = new unsigned char[16];

#ifdef __WINDOWS__
// Acquire an crypto provider - we use the basic one built into IE4 and above
	wxLogDebug(wxT("CreateFileHashED2k: acquiring MS basic crypto provider"));
	if (! (CryptAcquireContext(&CryptoProvider, NULL, MS_DEF_PROV, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT))) {
		SendError(ERR_HASH_CREATE);
		InputFile->Close();
		wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing buffers"));
		delete[] buffer;
		delete[] hashes;
		delete[] hash;
		wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashED2k: Error"));
		return;
	}

// Create the first level of hashes (the hashes for each DEF_ED2KBLOCK)
	wxLogDebug(wxT("CreateFileHashED2k: running level 1 hash"));
	for(cnt = 0; cnt < hashcnt; cnt++) {
		blockleft = DEF_ED2KBLOCK;
		if (!(CryptCreateHash(CryptoProvider, CALG_MD4, 0, 0, &CryptoHash))) {
			SendError(ERR_HASH_CREATE);
			InputFile->Close();
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing InputFile"));
			delete InputFile;
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing buffers"));
			delete[] buffer;
			delete[] hashes;
			delete[] hash;
			wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashED2k: Error"));
			return;
		}
		while (blockleft >= bufsize ) {
			read = InputFile->Read(buffer, bufsize);
			if (read == 0) blockleft = 0;
			if (! (CryptHashData(CryptoHash, buffer, read, 0))) {
				SendError(ERR_HASH_CREATE);
				InputFile->Close();
				wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing InputFile"));
				delete InputFile;
				wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing buffers"));
				delete[] buffer;
				delete[] hashes;
				delete[] hash;
				wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashED2k: Error"));
				return;
			}
			blockleft -= read;

// Update the progress bar
			SendProgress((wxInt8)((((DEF_ED2KBLOCK * (frUint64)(cnt + 1)) - blockleft) * 100) / InputFile->Length()));
		}
		if (blockleft > 0) {
			read = InputFile->Read(buffer, blockleft);
			if (! (CryptHashData(CryptoHash, buffer, read, 0))) {
				SendError(ERR_HASH_CREATE);
				InputFile->Close();
				wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing InputFile"));
				delete InputFile;
				wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing buffers"));
				delete[] buffer;
				delete[] hashes;
				delete[] hash;
				wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashED2k: Error"));
				return;
			}
// Update the progress bar
			SendProgress((wxInt8)(((DEF_ED2KBLOCK * (frUint64)(cnt + 1)) * 100) / InputFile->Length()));
		}

		hashlen= 16;
		if (! (CryptGetHashParam(CryptoHash, HP_HASHVAL, hash, &hashlen, 0))) {
			SendError(ERR_HASH_CREATE);
			InputFile->Close();
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing InputFile"));
			delete InputFile;
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing buffers"));
			delete[] buffer;
			delete[] hashes;
			delete[] hash;
			wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashED2k: Error"));
			return;
		}
// Append this hash to the end of the buffer ready for the second stage
		memcpy(hashes + (cnt * 16), hash, 16);
// Check whether the user has stopped the process
		if (CheckExit()) {
			InputFile->Close();
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing InputFile"));
			delete InputFile;
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing buffers"));
			delete[] buffer;
			delete[] hashes;
			delete[] hash;
			wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashED2k: Abort"));
			return;
		}
	}

// If we have more than one block then run the second stage hash.
// This creates a hash based on the results from our previous hashes.
	if (hashcnt > 1) {
		wxLogDebug(wxT("CreateFileHashED2k: running level 2 hash"));
		if (!(CryptCreateHash(CryptoProvider, CALG_MD4, 0, 0, &CryptoHash))) {
			SendError(ERR_HASH_CREATE);
			InputFile->Close();
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing InputFile"));
			delete InputFile;
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing buffers"));
			delete[] buffer;
			delete[] hashes;
			delete[] hash;
			wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashED2k: Error"));
			return;
		}
		if (! (CryptHashData(CryptoHash, hashes, hashcnt * 16, 0))) {
			SendError(ERR_HASH_CREATE);
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing InputFile"));
			delete InputFile;
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing buffers"));
			delete[] buffer;
			delete[] hashes;
			delete[] hash;
			wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashED2k: Error"));
			return;
		}
		hashlen = 16;
		if (! (CryptGetHashParam(CryptoHash, HP_HASHVAL, hash, &hashlen, 0))) {
			SendError(ERR_HASH_CREATE);
			InputFile->Close();
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing InputFile"));
			delete InputFile;
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing buffers"));
			delete[] buffer;
			delete[] hashes;
			delete[] hash;
			wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashED2k: Error"));
			return;
		}
#endif
#ifdef __UNIX__
// On UNIX we'll use the OpenSSL functions for MD4.
// No point in re-inventing the wheel.
	OpenSSL_add_all_digests();

// Create the first level of hashes (the hashes for each DEF_ED2KBLOCK)
	wxLogDebug(wxT("CreateFileHashED2k: running level 1 hash"));
	for (cnt = 0; cnt < hashcnt; cnt++) {
		blockleft = DEF_ED2KBLOCK;

		EVP_DigestInit(&md_temp, EVP_get_digestbyname("MD4"));

		while (blockleft >= bufsize) {
			read = InputFile->Read(buffer, bufsize);
			if (read == 0) blockleft = 0;
			EVP_DigestUpdate(&md_temp, buffer, read);
			blockleft -= read;
// Update the progress bar
			SendProgress((wxInt8)((((DEF_ED2KBLOCK * (frUint64)(cnt + 1)) - blockleft) * 100) / InputFile->Length()));
// Check whether the user has stopped the process
			if (CheckExit()) {
				InputFile->Close();
				wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing InputFile"));
				delete InputFile;
				wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing buffers"));
				delete[] buffer;
				delete[] hashes;
				delete[] hash;
				wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashED2k: Abort"));
				return;
			}
		}
		if (blockleft > 0) {
			read = InputFile->Read(buffer, blockleft);
			EVP_DigestUpdate(&md_temp, buffer, read);
// Update the progress bar
			SendProgress((wxInt8)(((DEF_ED2KBLOCK * (frUint64)(cnt + 1)) * 100) / InputFile->Length()));
		}

		hashlen = 16;
		EVP_DigestFinal(&md_temp, hash, &hashlen);
// Append this hash to the end of the buffer ready for the second stage
		memcpy(hashes + (cnt * 16), hash, 16);
// Check whether the user has stopped the process
		if (CheckExit()) {
			InputFile->Close();
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing InputFile"));
			delete InputFile;
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing buffers"));
			delete[] buffer;
			delete[] hashes;
			delete[] hash;
			wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashED2k: Abort"));
			return;
		}
	}

// If we have more than one block then run the second stage hash.
// This creates a hash based on the results from our previous hashes.
	if (hashcnt > 1) {
		wxLogDebug(wxT("CreateFileHashED2k: running level 2 hash"));
		EVP_DigestInit(&md_temp, EVP_get_digestbyname("MD4"));
		EVP_DigestUpdate(&md_temp, hashes, hashcnt * 16);
		hashlen = 16;
		EVP_DigestFinal(&md_temp, hash, &hashlen);
#endif
	} else {
		memcpy(hash, hashes, 16);
	}
	InputFile->Close();
// Update the progress bar
	SendProgress((wxInt8)100);
	SendStatus(wxT(STAT_DONE));
// Reset the progress bar
	SendProgress((wxInt8)0);

// Output the results.
// We take each byte and output it as a two-digit hex code.
	wxLogDebug(wxT("CreateFileHashED2k: outputting results"));
	message.Clear();
	for (cnt = 0; cnt < 16; cnt++)
		message += wxString::Format(wxT("%02x"), hash[cnt]).Right(2);
	message += wxString::Format(wxT(" *%s"), wxFileNameFromPath(FileName).c_str());

	SendHash(message);

	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing InputFile"));
	delete InputFile;
	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashED2k: freeing buffers"));
	delete[] buffer;
	delete[] hashes;
	delete[] hash;
	wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashED2k: Exit"));
	return;
}

//
// CreateFileHashMD5
//
// Receives a filename and buffer size as input.  Creates an MD5 hash
// from the file contents.  This is the same as md5sum does.
//
void FR_BackEnd::CreateFileHashMD5(const wxString FileName, const wxUint32 bufsize)
{
	wxString        message;
	frFFile         *InputFile;
	wxUint32        read;
	frUint64        cnt = 0;
	unsigned int    counter;
	unsigned char   *buffer,
	                *hash;
#ifdef __WINDOWS__
	unsigned long   hashlen;
	HCRYPTPROV      CryptoProvider;
	HCRYPTHASH      CryptoHash;
#endif
#ifdef __UNIX__
	wxUint32        hashlen;
	EVP_MD_CTX      md_temp;
#endif

	wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashMD5: Entry"));
	wxASSERT(bufsize > 0);
	wxASSERT(FileName.Len() > 0);
	SendStatus(wxString::Format(wxT(STAT_MD_CREATE), wxT("MD5"), FileName.c_str()));

// Open the input file
	wxLogDebug(wxString::Format(wxT("CreateFileHashMD5: opening file %s"), FileName.c_str()));
	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: allocating InputFile"));
	InputFile = new frFFile(FileName, wxT("rb"));
	if (! InputFile->IsOpened()) {
		SendError(wxString::Format(wxT(ERR_FILESTAT), FileName.c_str()));
		wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashMD5: Error"));
		return;
	}

	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: allocating buffers"));
	buffer = new unsigned char[bufsize];
	hash = new unsigned char[16];
	hashlen = 16;

#ifdef __WINDOWS__
// Acquire an crypto provider - we use the basic one built into IE4 and above
	wxLogDebug(wxT("CreateFileHashMD5: acquiring MS basic crypto provider"));
	if (! (CryptAcquireContext(&CryptoProvider, NULL, MS_DEF_PROV, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT))) {
		SendError(wxT(ERR_HASH_CREATE));
		InputFile->Close();
		wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: freeing buffers"));
		delete[] buffer;
		delete[] hash;
		wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashMD5: Error"));
		return;
	}

	if (!(CryptCreateHash(CryptoProvider, CALG_MD5, 0, 0, &CryptoHash))) {
		SendError(ERR_HASH_CREATE);
		InputFile->Close();
		wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: freeing buffers"));
		delete[] buffer;
		delete[] hash;
		wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashMD5: Error"));
		return;
	}

// Go through the input file and calculate the MD5
	wxLogDebug(wxT("CreateFileHashMD5: creating hash"));
	while(! InputFile->Eof()) {
		read = InputFile->Read(buffer, bufsize);
		if (read > 0) {
			if (! (CryptHashData(CryptoHash, buffer, read, 0))) {
				SendError(wxT(ERR_HASH_CREATE));
				InputFile->Close();
				wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: freeing InputFile"));
				delete InputFile;
				wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: freeing buffers"));
				delete[] buffer;
				delete[] hash;
				wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashMD5: Error"));
				return;
			}
// Update the progress bar
			SendProgress((wxInt8)((((frUint64)(cnt * bufsize) + read) * 100) / InputFile->Length()));
			cnt++;
		}
// Check whether the user has stopped the process
		if (CheckExit()) {
			InputFile->Close();
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: freeing InputFile"));
			delete InputFile;
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: freeing buffers"));
			delete[] buffer;
			delete[] hash;
			wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashMD5: Abort"));
			return;
		}
	}

	if (! (CryptGetHashParam(CryptoHash, HP_HASHVAL, hash, &hashlen, 0))) {
		SendError(ERR_HASH_CREATE);
		InputFile->Close();
		wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: freeing buffers"));
		delete[] buffer;
		delete[] hash;
		wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashMD5: Error"));
		return;
	}
#endif
#ifdef __UNIX__
// On UNIX we'll use the OpenSSL functions for MD5.
// No point in re-inventing the wheel.
	OpenSSL_add_all_digests();

// Go through the input file and calculate the MD5
	EVP_DigestInit(&md_temp, EVP_get_digestbyname("MD5"));
	wxLogDebug(wxT("CreateFileHashMD5: creating hash"));
	while(! InputFile->Eof()) {
		read = InputFile->Read(buffer, bufsize);
		if (read > 0) {
			EVP_DigestUpdate(&md_temp, buffer, read);
// Update the progress bar
			SendProgress((wxInt8)((((frUint64)(cnt * bufsize) + read) * 100) / InputFile->Length()));
			cnt++;
		}
// Check whether the user has stopped the process
		if (CheckExit()) {
			InputFile->Close();
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: freeing InputFile"));
			delete InputFile;
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: freeing buffers"));
			delete[] buffer;
			delete[] hash;
			wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashMD5: Abort"));
			return;
		}
	}

	EVP_DigestFinal(&md_temp, hash, &hashlen);

#endif
	InputFile->Close();
// Update the progress bar
	SendProgress((wxInt8)100);
	SendStatus(wxT(STAT_DONE));
// Reset the progress bar
	SendProgress((wxInt8)0);

// Output the results.
// We take each byte and output it as a two-digit hex code.
	wxLogDebug(wxT("CreateFileHashMD5: outputting results"));
	message.Clear();
	for (counter = 0; counter < 16; counter++)
		message += wxString::Format(wxT("%02x"), hash[counter]).Right(2);
	message += wxString::Format(wxT(" *%s"), wxFileNameFromPath(FileName).c_str());

	SendHash(message);

	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: freeing InputFile"));
	delete InputFile;
	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashMD5: freeing buffers"));
	delete[] buffer;
	delete[] hash;
	wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashMD5: Exit"));
	return;
}

//
// CreateFileHashCRC
//
// Receives a filename and buffer size as input.  Creates a CRC32 hash
// from the file contents.  This is the same as crc32/sfv does.
//
void FR_BackEnd::CreateFileHashCRC(const wxString FileName, const wxUint32 bufsize)
{
	wxString        message;
	frFFile         *InputFile;
	wxUint32        read;
	frUint64        cnt = 0;
	unsigned char   *buffer;
	tcrc            hash = INITCRC;

	wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashCRC: Entry"));
	wxASSERT(bufsize > 0);
	wxASSERT(FileName.Len() > 0);
	SendStatus(wxString::Format(wxT(STAT_MD_CREATE), wxT("CRC"), FileName.c_str()));

// Open the input file
	wxLogDebug(wxString::Format(wxT("CreateFileHashCRC: opening file %s"), FileName.c_str()));
	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashCRC: allocating InputFile"));
	InputFile = new frFFile(FileName, wxT("rb"));
	if (! InputFile->IsOpened()) {
		SendError(wxString::Format(wxT(ERR_FILESTAT), FileName.c_str()));
		wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashCRC: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashCRC: Error"));
		return;
	}

// Initialise the CRC table
	init_crctab(CRC_POLY);

	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashCRC: allocating buffers"));
	buffer = new unsigned char[bufsize];

// Go through the input file and calculate the CRC
	wxLogDebug(wxT("CreateFileHashCRC: creating hash"));
	while(! InputFile->Eof()) {
		read = InputFile->Read(buffer, bufsize);
		if (read > 0) {
			hash =  createcrc(buffer, read, hash);
// Update the progress bar
			SendProgress((wxInt8)((((frUint64)(cnt * bufsize) + read) * 100) / InputFile->Length()));
			cnt++;
		}
// Check whether the user has stopped the process
		if (CheckExit()) {
			InputFile->Close();
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashCRC: freeing InputFile"));
			delete InputFile;
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashCRC: freeing buffers"));
			delete[] buffer;
			wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashCRC: Abort"));
			return;
		}
	}

	InputFile->Close();
// Update the progress bar
	SendProgress((wxInt8)100);
	SendStatus(wxT(STAT_DONE));
// Reset the progress bar
	SendProgress((wxInt8)0);
	
	hash ^= 0xFFFFFFFFL;

// Output the results.
	wxLogDebug(wxT("CreateFileHashCRC: outputting results"));
	message = wxString::Format(wxT("%08X *%s"), (unsigned int)hash, wxFileNameFromPath(FileName).c_str());

	SendHash(message);

	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashCRC: freeing InputFile"));
	delete InputFile;
	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashCRC: freeing buffers"));
	delete[] buffer;
	wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashCRC: Exit"));
	return;
}

//
// CreateFileHashTTH
//
// Receives a filename and buffer size as input.  Creates a Tiger tree hash
// from the file contents (this is an Tiger based hash - each block
// from the file has a Tiger hash created for it.  All of these hashes are
// then appended and a tiger hash is created for this.
//
void FR_BackEnd::CreateFileHashTTH(const wxString FileName, const wxUint32 bufsize)
{
	frFFile         *InputFile;
	wxUint32        read;
	wxString        message;
	unsigned char   *buffer,
			hash[TIGERSIZE];
	frUint64        total;
	TT_CONTEXT	thash;

	wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashTTH: Entry"));
	wxASSERT(bufsize > 0);
	wxASSERT(FileName.Len() > 0);
	SendStatus(wxString::Format(wxT(STAT_MD_CREATE), wxT("TTH"), FileName.c_str()));

// Open the input file
	wxLogDebug(wxString::Format(wxT("CreateFileHashTTH: opening file %s"), FileName.c_str()));
	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashTTH: allocating InputFile"));
	InputFile = new frFFile(FileName, wxT("rb"));
	if (! InputFile->IsOpened()) {
		SendError(wxString::Format(wxT(ERR_FILESTAT), FileName.c_str()));
		wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashTTH: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashTTH: Error"));
		return;
	}

// Allocate buffer for file data
	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashTTH: allocating buffers"));
	buffer = new unsigned char[bufsize];

// Do the TigerTreeHash
	tt_init(&thash);
	total = 0;
	while ((read = InputFile->Read(buffer, bufsize))) {
		total += read;
		tt_update(&thash, buffer, read);
// Update the progress bar
		SendProgress((wxInt8)((frUint64)(total * 100) / InputFile->Length()));
// Check whether the user has stopped the process
		if (CheckExit()) {
			InputFile->Close();
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashTTH: freeing InputFile"));
			delete InputFile;
			wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashTTH: freeing buffers"));
			delete[] buffer;
			wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashTTH: Abort"));
			return;
		}
	}
// Create the output digest
	tt_digest(&thash, hash);

	InputFile->Close();
// Update the progress bar
	SendProgress((wxInt8)100);
	SendStatus(wxT(STAT_DONE));
// Reset the progress bar
	SendProgress((wxInt8)0);

// Output the results.
// We take each byte and output it as a two-digit hex code.
	wxLogDebug(wxT("CreateFileHashTTH: outputting results"));
	message.Clear();
	toBase32(hash, 24, message);
	message += wxString::Format(wxT(" *%s"), wxFileNameFromPath(FileName).c_str());

	SendHash(message);

	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashTTH: freeing InputFile"));
	delete InputFile;
	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateFileHashTTH: freeing buffers"));
	delete[] buffer;
	wxLogTrace(wxT("frTRACE_functions"), wxT("CreateFileHashTTH: Exit"));
	return;
}

//
// CreateCRC
//
// Receives a filename, checkfile name and buffer size as input.  Creates
// a checkfile containing a list of CRCs (one per buffer size block) from
// the file contents.
//
void FR_BackEnd::CreateCRC(const wxString FileName, const wxString ChkFile, const wxUint32 bufsize)
{
	wxTextFile      CheckFile;
	frFFile         *InputFile;
	CRCList         *localcrc;
	CRCList::Node   *node;

	wxLogTrace(wxT("frTRACE_functions"), wxT("CreateCRC: Entry"));
	wxASSERT(bufsize > 0);
	wxASSERT(ChkFile.Len() > 0);
	wxASSERT(FileName.Len() > 0);
// Check that the checkfile does not already exist
	wxLogDebug(wxString::Format(wxT("CreateCRC: creating whether %s already exists"), ChkFile.c_str()));
	if (wxFileExists(ChkFile)) {
		SendError(wxString::Format(wxT(ERR_FILE_EXISTS), ChkFile.c_str()));
		wxLogTrace(wxT("frTRACE_functions"), wxT("CreateCRC: Error"));
		return;
	}
// Create a list of CRCs
	wxLogDebug(wxT("CreateCRC: creating CRCs"));
	localcrc = crclist(FileName, bufsize);
	if (localcrc == NULL) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("CreateCRC: Error"));
		return;
	}
// Create the output file
	wxLogDebug(wxString::Format(wxT("CreateCRC: creating file %s"), FileName.c_str()));
	if (! CheckFile.Create(ChkFile)) {
		SendError(wxString::Format(wxT(ERR_FILE_IO), ChkFile.c_str(), wxSysErrorMsg()));
		wxLogTrace(wxTRACE_MemAlloc, wxT("CreateCRC: freeing localcrc"));
		localcrc->DeleteContents(TRUE);
		wxLogTrace(wxT("frTRACE_functions"), wxT("CreateCRC: Error"));
		return;
	}
// Write out all the header information (program name and version, blocksize
// used and size of input file.
	wxLogDebug(wxT("CreateCRC: writing header info to file"));
	CheckFile.AddLine(wxString::Format(wxT("FILEREPAIR %s"), wxT(VERSION)));
	CheckFile.AddLine(wxString::Format(wxT("BLOCKSIZE %d"), bufsize/1024));
	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateCRC: allocating InputFile"));
	InputFile = new frFFile(FileName, wxT("rb"));
	CheckFile.AddLine(wxString::Format(wxT("FILESIZE %" wxLongLongFmtSpec "d"), (frInt64)InputFile->Length()));
	InputFile->Close();
	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateCRC: freeing InputFile"));
	delete InputFile;

// Write out the list of CRCs, one per line
	wxLogDebug(wxT("CreateCRC: writing CRCs to file"));
	node = localcrc->GetFirst();
	while (node) {
		CheckFile.AddLine(wxString::Format(wxT("%010lu"), *(node->GetData())));
		node = node->GetNext();
	}
	CheckFile.Write();
	CheckFile.Close();
	wxLogTrace(wxTRACE_MemAlloc, wxT("CreateCRC: freeing localcrc"));
	localcrc->DeleteContents(TRUE);
	SendStatus(wxT(STAT_DONE));
// Reset the progress bar
	SendProgress((wxInt8)0);
	wxLogTrace(wxT("frTRACE_functions"), wxT("CreateCRC: Exit"));
	return;
}

//
// VerifyCRC
//
// Receives a filename and checkfile name input.  It verifies whether the
// CRCs in the checkfile match those it creates for the input file and
// counts the number of mismatches.
//
void FR_BackEnd::VerifyCRC(const wxString FileName, const wxString ChkFile)
{
	wxTextFile      CheckFile;
	frFFile         *InputFile;
	wxString        versionin,
	                temp;
	unsigned long   blocksize;
	frInt64         filesize;
	wxUint32        errcnt;
	CRCList         *localcrc,
	                remotecrc;
	ErrorList       errors;
	tcrc            *crcin;


	wxLogTrace(wxT("frTRACE_functions"), wxT("VerifyCRC: Entry"));
	wxASSERT(ChkFile.Len() > 0);
	wxASSERT(FileName.Len() > 0);
// Open the checkfile
	wxLogDebug(wxString::Format(wxT("VerifyCRC: opening checkfile %s"), ChkFile.c_str()));
	if (! CheckFile.Open(ChkFile)) {
		SendError(wxString::Format(wxT(ERR_FILE_IO), ChkFile.c_str(), wxSysErrorMsg()));
		wxLogTrace(wxT("frTRACE_functions"), wxT("VerifyCRC: Error"));
		return;
	}
// Check the header information (verify that the program is the same, read in
// the blocksize to use and verify the filesizes match)
	wxLogDebug(wxT("VerifyCRC: checking header info"));
	if (! CheckFile.GetFirstLine().StartsWith(wxT("FILEREPAIR "), &versionin)) {
		SendError(wxString::Format(wxT(ERR_CHKFILE), wxT("version"), ChkFile.c_str()));
		CheckFile.Close();
		wxLogTrace(wxT("frTRACE_functions"), wxT("VerifyCRC: Error"));
		return;
	}
	if (! CheckFile.GetNextLine().StartsWith(wxT("BLOCKSIZE "), &temp)) {
		SendError(wxString::Format(wxT(ERR_CHKFILE), wxT("blocksize"), ChkFile.c_str()));
		CheckFile.Close();
		wxLogTrace(wxT("frTRACE_functions"), wxT("VerifyCRC: Error"));
		return;
	}
	if (! temp.ToULong(&blocksize)) blocksize = 0;
	if (blocksize == 0) {
		SendError(wxT(ERR_INVALID_BLOCK));
		CheckFile.Close();
		wxLogTrace(wxT("frTRACE_functions"), wxT("VerifyCRC: Error"));
		return;
	}
	if (! CheckFile.GetNextLine().StartsWith(wxT("FILESIZE "), &temp)) {
		SendError(wxString::Format(wxT(ERR_CHKFILE), wxT("filesize"), ChkFile.c_str()));
		CheckFile.Close();
		wxLogTrace(wxT("frTRACE_functions"), wxT("VerifyCRC: Error"));
		return;
	}
	filesize = to_int64(temp);
	if (filesize == 0) {
		SendError(wxString::Format(wxT(ERR_CHKFILE), wxT("filesize"), ChkFile.c_str()));
		CheckFile.Close();
		wxLogTrace(wxT("frTRACE_functions"), wxT("VerifyCRC: Error"));
		return;
	}

	wxLogDebug(wxString::Format(wxT("VerifyCRC: opening file %s"), FileName.c_str()));
	wxLogTrace(wxTRACE_MemAlloc, wxT("VerifyCRC: allocating InputFile"));
	InputFile = new frFFile(FileName, wxT("rb"));
	if (InputFile->Length() != filesize) {
		SendError(wxString::Format(wxT(ERR_FILESIZE1), filesize) + wxString::Format(wxT(ERR_FILESIZE2), (frInt64)InputFile->Length()));
		CheckFile.Close();
		InputFile->Close();
		wxLogTrace(wxTRACE_MemAlloc, wxT("VerifyCRC: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxT("frTRACE_functions"), wxT("VerifyCRC: Error"));
		return;
	}
	InputFile->Close();
	wxLogTrace(wxTRACE_MemAlloc, wxT("VerifyCRC: freeing InputFile"));
	delete InputFile;

// Read in the CRCs from the checkfile into a list
	wxLogDebug(wxT("VerifyCRC: reading CRCs into list"));
	blocksize = blocksize * 1024;
	while (! CheckFile.Eof()) {
		wxLogTrace(wxTRACE_MemAlloc, wxT("VerifyCRC: allocating crcin"));
		crcin = new tcrc();
		if (CheckFile.GetNextLine().ToULong(crcin))
			remotecrc.Append(crcin);
		else
			delete crcin;
	}
	CheckFile.Close();

// Create a list of the CRCs for our input file
	wxLogDebug(wxT("VerifyCRC: creating local CRC list"));
	localcrc = crclist(FileName, blocksize);
	if (localcrc == NULL) {
		wxLogTrace(wxTRACE_MemAlloc, wxT("VerifyCRC: freeing remotecrc"));
		remotecrc.DeleteContents(TRUE);
		wxLogTrace(wxT("frTRACE_functions"), wxT("VerifyCRC: Error"));
		return;
	}
// Compare the two lists and get a count of mismatches
	wxLogDebug(wxT("VerifyCRC: comparing CRC lists"));
	if (! crc_compare(&remotecrc, localcrc, &errors)) {
		wxLogTrace(wxTRACE_MemAlloc, wxT("VerifyCRC: freeing remotecrc"));
		remotecrc.DeleteContents(TRUE);
		wxLogTrace(wxTRACE_MemAlloc, wxT("VerifyCRC: freeing localcrc"));
		localcrc->DeleteContents(TRUE);
		delete localcrc;
		wxLogTrace(wxT("frTRACE_functions"), wxT("VerifyCRC: Error"));
		return;
	}
	errcnt = errors.GetCount();
	SendStatus(wxT(STAT_DONE));
// Reset the progress bar
	SendProgress((wxInt8)0);

	wxLogTrace(wxTRACE_MemAlloc, wxT("VerifyCRC: freeing errors"));
	errors.DeleteContents(TRUE);
	wxLogTrace(wxTRACE_MemAlloc, wxT("VerifyCRC: freeing remotecrc"));
	remotecrc.DeleteContents(TRUE);
	wxLogTrace(wxTRACE_MemAlloc, wxT("VerifyCRC: freeing localcrc"));
	localcrc->DeleteContents(TRUE);

// Report the number of errors found and, if there were errors, the blocksize
// that was used.
	wxLogDebug(wxT("VerifyCRC: reporting results"));
	if (errcnt > 0)
		SendAlert(wxString::Format(wxT(ALRT_ERR_CNT), errcnt, (wxUint32)blocksize/1024));
	else
		SendAlert(wxString(wxT(ALRT_ERR_NONE)));

	wxLogTrace(wxT("frTRACE_functions"), wxT("VerifyCRC: Exit"));
	return;
}

//
// Client
//
// Receives a filename, server name, server port, blocksize and flag for
// passive mode as input.  It connects to the server on the given port and
// then, depending on the passive mode flag, continues as the client or
// switches into server mode.
//
void FR_BackEnd::Client(const wxString FileName, const wxString server, const unsigned short port, const wxUint32 bufsize, const bool passive)
{
	wxSocketClient  s;
	wxIPV4address   srv_addr;
	frFFile         *InputFile;
	bool            retval;

	wxLogTrace(wxT("frTRACE_functions"), wxT("Client: Entry"));
	wxASSERT(FileName.Len() > 0);
	wxASSERT(server.Len() > 0);
	wxASSERT(port > 0);
	wxASSERT(bufsize > 0);
	oldversion = TRUE;

// Open the input file
	wxLogDebug(wxString::Format(wxT("Client: opening file %s"), FileName.c_str()));
	wxLogTrace(wxTRACE_MemAlloc, wxT("Client: allocating InputFile"));
	if (passive)
		InputFile = new frFFile(FileName, wxT("rb"));
	else
		InputFile = new frFFile(FileName, wxT("rb+"));

	if (! InputFile->IsOpened()) {
		SendError(wxString::Format(wxT(ERR_FILE_IO), FileName.c_str(), wxSysErrorMsg()));
		wxLogTrace(wxTRACE_MemAlloc, wxT("Client: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxT("frTRACE_functions"), wxT("Client: Error"));
		return;
	}

// Set up the server details
	if ((! srv_addr.Hostname(server)) || (! srv_addr.Service(port))) {
		SendError(wxString::Format(wxT(ERR_SERVER), server.c_str()));
		InputFile->Close();
		wxLogTrace(wxTRACE_MemAlloc, wxT("Client: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxT("frTRACE_functions"), wxT("Client: Error"));
		return;
	}

// Connect to the server
	wxLogDebug(wxString::Format(wxT("Client: connecting to server %s on port %d"), server.c_str(), port));
	if ((! s.Connect(srv_addr)) || (! s.IsConnected())) {
		SendError(wxT(ERR_NO_SERVER));
		InputFile->Close();
		wxLogTrace(wxTRACE_MemAlloc, wxT("Client: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxT("frTRACE_functions"), wxT("Client: Error"));
		return;
	}

	SendStatus(wxString::Format(wxT(STAT_C_CONN), server.c_str()));

// Set the socket not to wait until all information has been sent
// or received.
	s.SetFlags(wxSOCKET_NOWAIT);

// Check the passive mode flag and continue in the required mode
	if (passive) {
		wxLogDebug(wxT("Client: switching to server mode"));
		retval = server_mode(&s, InputFile, FileName);
	} else {
		wxLogDebug(wxT("Client: continuing in client mode"));
		retval = client_mode(&s, InputFile, bufsize, FileName);
	}

	InputFile->Close();
	wxLogTrace(wxTRACE_MemAlloc, wxT("Client: freeing InputFile"));
	delete InputFile;
	s.Close();
	if (retval) SendStatus(wxT(STAT_DONE));
// Reset the progress bar
	SendProgress((wxInt8)0);
	wxLogTrace(wxT("frTRACE_functions"), wxT("Client: Exit"));
	return;
}

//
// Server
//
// Receives a filename, listening port and flag for passive mode as input.
// It then starts up listening for a client connection on the given port.
// Once a client has connected it decides, based on the passive flag,
// whether to continue as a server or to switch into client mode.
//
void FR_BackEnd::Server(const wxString FileName, const unsigned short port, const bool passive) {
	wxSocketServer  *s;
	wxSocketBase    *ns = NULL;
	wxIPV4address   srv_addr;
	frFFile         *InputFile;
	bool            retval;

	wxLogTrace(wxT("frTRACE_functions"), wxT("Server: Entry"));
	wxASSERT(FileName.Len() > 0);
	wxASSERT(port > 0);
	oldversion = TRUE;

// Open the input file
	wxLogDebug(wxString::Format(wxT("Server: opening file %s"), FileName.c_str()));
	wxLogTrace(wxTRACE_MemAlloc, wxT("Server: allocating InputFile"));
	if (passive)
		InputFile = new frFFile(FileName, wxT("rb+"));
	else
		InputFile = new frFFile(FileName, wxT("rb"));

	if(! InputFile->IsOpened()) {
		SendError(wxString::Format(wxT(ERR_FILE_IO), FileName.c_str(), wxSysErrorMsg()));
		wxLogTrace(wxTRACE_MemAlloc, wxT("Server: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxT("frTRACE_functions"), wxT("Server: Error"));
		return;
	}

// Set up the server details
	if (! srv_addr.Service(port)) {
		NetError(wxSOCKET_INVPORT);
		InputFile->Close();
		wxLogTrace(wxTRACE_MemAlloc, wxT("Server: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxT("frTRACE_functions"), wxT("Server: Error"));
		return;
	}

// Create the socket
	wxLogTrace(wxTRACE_MemAlloc, wxT("Server: allocating s"));
	s = new wxSocketServer(srv_addr);

	if (! s->Ok()) {
		NetError(s->LastError());
		wxLogTrace(wxTRACE_MemAlloc, wxT("Server: freeing s"));
		delete s;
		InputFile->Close();
		wxLogTrace(wxTRACE_MemAlloc, wxT("Server: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxT("frTRACE_functions"), wxT("Server: Error"));
		return;
	}

	SendStatus(wxString::Format(wxT(STAT_S_LISTEN), FileName.c_str()));

// Start listening on the port.  Check every second for a pending connection.
// This means we can still check for the user stopping the server as well.
	wxLogDebug(wxString::Format(wxT("Server: listening on port %d"), port));
	do {
// Check whether the user has stopped the process
		if (CheckExit()) {
			wxLogTrace(wxTRACE_MemAlloc, wxT("Server: freeing s"));
			s->Destroy();
			InputFile->Close();
			wxLogTrace(wxTRACE_MemAlloc, wxT("Server: freeing InputFile"));
			delete InputFile;
			wxLogTrace(wxT("frTRACE_functions"), wxT("Server: Abort"));
			return;
		}
		wxLogTrace(wxT("frTRACE_net"), wxT("Server: checking connection"));
		ns = s->Accept(FALSE);
		if (ns == NULL)
			Sleep(1000);
	} while (ns == NULL);
// We've now got a connection

// Once the connection is made close the original socket (we have a new socket
// for the open connection) as we don't want to accept any more connections.
	wxLogTrace(wxTRACE_MemAlloc, wxT("Server: freeing s"));
	s->Destroy();

	SendStatus(wxT(STAT_S_CONN));

// Set the socket not to wait for data to be sent/received
	ns->SetFlags(wxSOCKET_NOWAIT);

// Check which mode to continue in
	if (passive) {
		wxLogDebug(wxT("Server: switching to client mode"));
		retval = client_mode(ns, InputFile, DEF_BLOCKSIZE, FileName);
	} else {
		wxLogDebug(wxT("Server: continuing in server mode"));
		retval = server_mode(ns, InputFile, FileName);
	}

	InputFile->Close();
	wxLogTrace(wxTRACE_MemAlloc, wxT("Server: freeing InputFile"));
	delete InputFile;
	wxLogTrace(wxTRACE_MemAlloc, wxT("Server: freeing ns"));
	ns->Destroy();
	if (retval) SendStatus(wxT(STAT_DONE));
// Reset the progress bar
	SendProgress((wxInt8)0);

	wxLogTrace(wxT("frTRACE_functions"), wxT("Server: Exit"));
	return;
}

//
// crclist
//
// Receives a filename and buffer size as input.  Creates a CRCList
// (a linked list of CRCs) based on the contents of the input file.
// Returns a pointer to this list (or NULL on error)
//
CRCList *FR_BackEnd::crclist(const wxString FileName, const wxUint32 bufsize)
{
	unsigned char   *buffer;
	frFFile         *InputFile;
	wxUint32        read;
	frUint64        cnt = 0;
	CRCList         *newlist;

	wxLogTrace(wxT("frTRACE_functions"), wxT("crclist: Entry"));
	wxASSERT(FileName.Len() > 0);
	wxASSERT(bufsize > 0);
	SendStatus(wxString::Format(wxT(STAT_CRC_CREATE_CRC), FileName.c_str()));

// Open the input file
	wxLogTrace(wxTRACE_MemAlloc, wxT("crclist: allocating InputFile"));
	InputFile = new frFFile(FileName, wxT("rb"));
	if (! InputFile->IsOpened()) {
		SendError(wxString::Format(wxT(ERR_FILE_IO), FileName.c_str(), wxSysErrorMsg()));
		wxLogTrace(wxTRACE_MemAlloc, wxT("crclist: freeing InputFile"));
		delete InputFile;
		wxLogTrace(wxT("frTRACE_functions"), wxT("crclist: Error"));
		return NULL;
	}
// Initialise the CRC table
	init_crctab(CRC_POLY);
// Create an empty CRCList
	wxLogTrace(wxTRACE_MemAlloc, wxT("crclist: allocating newlist"));
	newlist = new CRCList();
	wxLogTrace(wxTRACE_MemAlloc, wxT("crclist: allocating buffer"));
	buffer = new unsigned char[bufsize];
// For each buffer sized block, read it into the buffer and then
// append the calculated CRC for this block to the end of our list
	while(! InputFile->Eof()) {
		read = InputFile->Read(buffer, bufsize);
		if (read > 0) {
			newlist->Append(new tcrc(createcrc(buffer, read, INITCRC)));
// Update the progress bar
			SendProgress((wxInt8)((((frUint64)(cnt * bufsize) + read) * 100) / InputFile->Length()));
			cnt++;
		}
		if (CheckExit()) {
			InputFile->Close();
			wxLogTrace(wxTRACE_MemAlloc, wxT("crclist: freeing InputFile"));
			delete InputFile;
			wxLogTrace(wxTRACE_MemAlloc, wxT("crclist: freeing buffer"));
			delete[] buffer;
			wxLogTrace(wxTRACE_MemAlloc, wxT("crclist: freeing newlist"));
			newlist->DeleteContents(TRUE);
			wxLogTrace(wxT("frTRACE_functions"), wxT("crclist: Abort"));
			return NULL;
		}
	}
	InputFile->Close();
	wxLogTrace(wxTRACE_MemAlloc, wxT("crclist: freeing InputFile"));
	delete InputFile;
	wxLogTrace(wxTRACE_MemAlloc, wxT("crclist: freeing buffer"));
	delete[] buffer;
// Reset the progress bar
	SendProgress((wxInt8)0);
	wxLogTrace(wxT("frTRACE_functions"), wxT("crclist: Exit"));
	return newlist;
}

//
// createcrc
//
// Receives a buffer and size as inputs.  This code was grabbed from
// another program (which had in turn grabbed it from elsewhere).  The
// original copyright and explanations are included.  It creates and
// returns a 32-bit CRC for the provided size byte buffer.
//
tcrc FR_BackEnd::createcrc(const unsigned char *buffer, const wxUint32 size, tcrc crccode)
{
/*
 * Copyright (C) 1986 Gary S. Brown.  You may use this program, or
 * code or tables extracted from it, as desired without restriction.
 */
/* First, the polynomial itself and its table of feedback terms.  The  */
/* polynomial is                                                       */
/* X^32+X^26+X^23+X^22+X^16+X^12+X^11+X^10+X^8+X^7+X^5+X^4+X^2+X^1+X^0 */
/* Note that we take it "backwards" and put the highest-order term in  */
/* the lowest-order bit.  The X^32 term is "implied"; the LSB is the   */
/* X^31 term, etc.  The X^0 term (usually shown as "+1") results in    */
/* the MSB being 1.                                                    */

/* Note that the usual hardware shift register implementation, which   */
/* is what we're using (we're merely optimizing it by doing eight-bit  */
/* chunks at a time) shifts bits into the lowest-order term.  In our   */
/* implementation, that means shifting towards the right.  Why do we   */
/* do it this way?  Because the calculated CRC must be transmitted in  */
/* order from highest-order term to lowest-order term.  UARTs transmit */
/* characters in order from LSB to MSB.  By storing the CRC this way,  */
/* we hand it to the UART in the order low-byte to high-byte; the UART */
/* sends each low-bit to hight-bit; and the result is transmission bit */
/* by bit from highest- to lowest-order term without requiring any bit */
/* shuffling on our part.  Reception works similarly.                  */

/* The feedback terms table consists of 256, 32-bit entries.  Notes:   */
/*                                                                     */
/*     It might not be obvious, but the feedback terms simply	       */
/*     represent the results of eight shift/xor operations for all     */
/*     combinations of data and CRC register values.                   */
/*                                                                     */
/*     The values must be right-shifted by eight bits by the "updcrc"  */
/*     logic; the shift must be unsigned (bring in zeroes).  On some   */
/*     hardware you could probably optimize the shift in assembler by  */
/*     using byte-swap instructions.                                   */
// The table has been moved to a separate function as there's no need to
// recreate it for each bufferload.

	wxLogTrace(wxT("frTRACE_functions"), wxT("createcrc: Entry"));
	wxASSERT(size > 0);
	wxUint32        i;

	for (i = 0;  i < size;  i ++) {
		crccode = crctab[(wxUint32) ((crccode) ^ (buffer[i])) & 0xff] ^ (((crccode) >> 8) & 0x00FFFFFFL);
	}

	wxLogTrace(wxT("frTRACE_functions"), wxT("createcrc: Exit"));
	return crccode;
}

//
// crc_compare
//
// Receives two CRC lists and an error list as inputs.  It then compares the
// two CRC lists and, for each difference it finds, stores the offset of
// that difference in the errors list.
//
bool FR_BackEnd::crc_compare(CRCList *correct, CRCList *check, ErrorList *errors)
{
	wxUint32        count;
	CRCList::Node   *node,
	                *chknode;

	wxLogTrace(wxT("frTRACE_functions"), wxT("crc_compare: Entry"));
	SendStatus(wxT(STAT_CRC_COMP_CRC));

// Check the two CRC lists are of the same size
	if (correct->GetCount() != check->GetCount()) {
		SendError(wxT(ERR_FILESIZE1));
		wxLogTrace(wxT("frTRACE_functions"), wxT("crc_compare: Error"));
		return FALSE;
	}
// Compare values stored at the same offsets within the two lists.
// If the values differ then append that offset to the errors list.
	count = 0;
	chknode = check->GetFirst();
	for (node = correct->GetFirst(); node; node = node->GetNext()) {
		if (! chknode) {
			SendError(wxT(ERR_FILESIZE1));
			wxLogTrace(wxT("frTRACE_functions"), wxT("crc_compare: Error"));
			return FALSE;
		}
		if (*(node->GetData()) != *(chknode->GetData())) {
			wxLogTrace(wxTRACE_MemAlloc, wxT("crc_compare: allocating error"));
			errors->Append(new wxUint32(count));
		}
		SendProgress((wxInt8)((count * 100) / correct->GetCount()));
		count++;
		chknode = chknode->GetNext();
		if (CheckExit()) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("crc_compare: Abort"));
			return FALSE;
		}
	}
// Reset the progress bar
	SendProgress((wxInt8)0);
	wxLogTrace(wxT("frTRACE_functions"), wxT("crc_compare: Exit"));
	return TRUE;
}

//
// server_mode
//
// Receives an open socket, an opened input file and the name of that file
// as inputs.  It then manages the comparison of CRC lists and sending of
// any required blocks to the client.
//
bool FR_BackEnd::server_mode(wxSocketBase *sock, frFFile *infile, const wxString filename)
{
	wxUint32        blocksize;
	wxInt32         errcnt;
	CRCList         *localcrc,
	                remotecrc;
	ErrorList       errors;

	wxLogTrace(wxT("frTRACE_functions"), wxT("server_mode: Entry"));
// Check the client version
	wxLogDebug(wxT("server_mode: checking client version"));
	if (! check_version(sock)) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("server_mode: Error"));
		return FALSE;
	}

// Get the blocksize to use from the client
	wxLogDebug(wxT("server_mode: getting client blocksize"));
	blocksize = get_blocksize(sock);
	if (blocksize == 0) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("server_mode: Error"));
		return FALSE;
	}

// Check the client's filesize is the same as the local filesize
	wxLogDebug(wxT("server_mode: checking client filesize"));
	if (! check_filesize(sock, infile->Length())) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("server_mode: Error"));
		return FALSE;
	}

// Check whether the user has stopped the process
	if (CheckExit()) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("server_mode: Abort"));
		return FALSE;
	}

// Create a CRC list from the local file
	wxLogDebug(wxT("server_mode: creating local CRC list"));
	localcrc = crclist(filename, blocksize);

	if (localcrc == NULL) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("server_mode: Error"));
		return FALSE;
	}
// Check whether the user has stopped the process
	if (CheckExit()) {
		if (localcrc != NULL) {
			wxLogTrace(wxTRACE_MemAlloc, wxT("server_mode: freeing localcrc"));
			localcrc->DeleteContents(TRUE);
		}
		wxLogTrace(wxT("frTRACE_functions"), wxT("server_mode: Abort"));
		return FALSE;
	}
	SendStatus(wxT(STAT_S_COMP_CRC_DONE));

// Receive the client's CRC listings
	wxLogDebug(wxT("server_mode: getting client CRC list"));
	if (! get_client_crcs(sock, &remotecrc, infile->Length(), blocksize)) {
		wxLogTrace(wxTRACE_MemAlloc, wxT("server_mode: freeing localcrc"));
		localcrc->DeleteContents(TRUE);
		wxLogTrace(wxT("frTRACE_functions"), wxT("server_mode: Error"));
		return FALSE;
	}

// Compare the two lists and count the number of differences
	wxLogDebug(wxT("server_mode: comparing CRC lists"));
	errcnt = get_errcnt(sock, localcrc, &remotecrc, &errors, blocksize);

// If this is zero then there were no differences so stop now
	if (errcnt == 0) {
		wxLogTrace(wxTRACE_MemAlloc, wxT("server_mode: freeing localcrc"));
		localcrc->DeleteContents(TRUE);
		wxLogTrace(wxT("frTRACE_functions"), wxT("server_mode: Exit"));
		return FALSE;
	}
// A negative value means we have an error so abort here
	if (errcnt < 0) {
		if (errors.GetCount() > 0) {
			wxLogTrace(wxTRACE_MemAlloc, wxT("server_mode: freeing errors"));
			errors.DeleteContents(TRUE);
		}
		wxLogTrace(wxTRACE_MemAlloc, wxT("server_mode: freeing localcrc"));
		localcrc->DeleteContents(TRUE);
		wxLogTrace(wxT("frTRACE_functions"), wxT("server_mode: Error"));
		return FALSE;
	}

// Send any needed segments to the client
	wxLogDebug(wxT("server_mode: sending corrected segments"));
	if (! send_segments(sock, localcrc, &errors, blocksize, infile)) {
		wxLogTrace(wxTRACE_MemAlloc, wxT("server_mode: freeing errors"));
		errors.DeleteContents(TRUE);
		wxLogTrace(wxTRACE_MemAlloc, wxT("server_mode: freeing localcrc"));
		localcrc->DeleteContents(TRUE);
		wxLogTrace(wxT("frTRACE_functions"), wxT("server_mode: Error"));
		return FALSE;
	}
	wxLogTrace(wxTRACE_MemAlloc, wxT("server_mode: freeing errors"));
	errors.DeleteContents(TRUE);
	wxLogTrace(wxTRACE_MemAlloc, wxT("server_mode: freeing localcrc"));
	localcrc->DeleteContents(TRUE);
	wxLogTrace(wxT("frTRACE_functions"), wxT("server_mode: Exit"));
	return TRUE;
}

//
// client_mode
//
// Receives an open socket, an opened input file, the blocksize to use and
// the name of the input file.  It then manages the creation of the CRC list,
// sending it to the server and receiving any needed blocks.
//
bool FR_BackEnd::client_mode (wxSocketBase *sock, frFFile *infile, const wxUint32 blocksize, const wxString FileName)
{
	wxInt32 errcnt;
	CRCList *localcrc;

	wxLogTrace(wxT("frTRACE_functions"), wxT("client_mode: Entry"));
// Send the program name & version to the server
	wxLogDebug(wxT("client_mode: sending version"));
	if (! send_version(sock)) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("client_mode: Error"));
		return FALSE;
	}
// Send the blocksize to use for CRCs to the server
	wxLogDebug(wxT("client_mode: sending blocksize"));
	if (! send_blocksize(sock, blocksize)) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("client_mode: Error"));
		return FALSE;
	}
// Send the local filesize to the server
	wxLogDebug(wxT("client_mode: sending filesize"));
	if (! send_filesize(sock, infile)) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("client_mode: Error"));
		return FALSE;
	}
// Check whether the user has stopped the process
	if (CheckExit()) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("client_mode: Abort"));
		return FALSE;
	}
// Create the CRC list for the local file
	wxLogDebug(wxT("client_mode: creating local CRC list"));
	localcrc = crclist(FileName, blocksize);

	if (localcrc == NULL) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("client_mode: Error"));
		return FALSE;
	}
// Check whether the user has stopped the process
	if (CheckExit()) {
		if (localcrc != NULL) {
			wxLogTrace(wxTRACE_MemAlloc, wxT("client_mode: freeing localcrc"));
			localcrc->DeleteContents(TRUE);
		}
		wxLogTrace(wxT("frTRACE_functions"), wxT("client_mode: Abort"));
		return FALSE;
	}
// Send this list of CRCs to the server
	wxLogDebug(wxT("client_mode: sending CRC list to server"));
	if (! send_hashes (sock, blocksize, localcrc, infile->Length())) {
		wxLogTrace(wxTRACE_MemAlloc, wxT("client_mode: freeing localcrc"));
		localcrc->DeleteContents(TRUE);
		wxLogTrace(wxT("frTRACE_functions"), wxT("client_mode: Error"));
		return FALSE;
	}
	wxLogTrace(wxTRACE_MemAlloc, wxT("client_mode: freeing localcrc"));
	localcrc->DeleteContents(TRUE);

// Receive a count of incorrect blocks from the server
	wxLogDebug(wxT("client_mode: getting count of errors"));
	errcnt = rec_errcnt(sock);
	if (errcnt == 0) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("client_mode: Exit"));
		return FALSE;
	}
	if (errcnt < 1) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("client_mode: Error"));
		return FALSE;
	}
// Check whether the user has stopped the process
	if (CheckExit()) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("client_mode: Abort"));
		return FALSE;
	}

// Get correct versions of these blocks from the server
	wxLogDebug(wxT("client_mode: getting correct blocks"));
	receive_blocks (sock, blocksize, (wxUint32)errcnt, infile);
	wxLogTrace(wxT("frTRACE_functions"), wxT("client_mode: Exit"));
	return TRUE;
}

//
// SendLine
//
// Receives an open socket, a send buffer and the size of the buffer as
// inputs.  Send the buffer to the server.
//
bool FR_BackEnd::SendLine(wxSocketBase *sock, const unsigned char *buf, const wxUint32 len)
{
	wxUint32 length = wxUINT32_SWAP_ON_LE(len),
	         to_write;
	wxString len_str;

// Start by sending the buffer size.  On old versions this is sent as a
// string (in big-endian format - I got confused at the time okay!).  This
// version (and future versions) send it as an integer (again in big-endian
// format as this is standard for integers over the network).
	wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Entry"));
	wxLogTrace(wxT("frTRACE_net"), wxString::Format(wxT("SendLine: Length = %010u (%u), Version = %u"), length, len, oldversion));
	
	if (oldversion) {
		len_str.Printf(wxT("%010u)"), length);
		to_write = UINT32_LEN;
		wxLogTrace(wxT("frTRACE_net"), wxString::Format(wxT("SendLine: Sent %010u"), length));
	} else {
		to_write = sizeof(wxUint32);
		wxLogTrace(wxT("frTRACE_net"), wxString::Format(wxT("SendLine: Sent %u"), length));
	}
	
	while (to_write > 0) {
		if (oldversion)
			sock->Write(len_str.mb_str() + UINT32_LEN - to_write, to_write);
		else
			sock->Write(&length + sizeof(wxUint32) - to_write, to_write);

		if (sock->IsDisconnected()) {
			NetError(wxSOCKET_IOERR);
			wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Error"));
			return FALSE;
		}
		if (CheckExit()) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Abort"));
			return FALSE;
		}
		if ((sock->Error()) && (sock->LastError() != wxSOCKET_WOULDBLOCK)) {
			NetError(sock->LastError());
			wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Error"));
			return FALSE;
		}
		if (sock->LastCount() > 0)
			to_write = to_write - sock->LastCount();
		else
			Sleep(500);
	}


// Now we send the buffer itself.  Because we have told the socket to wait
// we don't need to check whether everything has been sent.
	to_write = len;
	while (to_write > 0) {
		sock->Write(buf + len - to_write, to_write);
		if (sock->IsDisconnected()) {
			NetError(wxSOCKET_IOERR);
			wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Error"));
			return FALSE;
		}
		if (CheckExit()) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Abort"));
			return FALSE;
		}
		if ((sock->Error()) && (sock->LastError() != wxSOCKET_WOULDBLOCK)) {
			NetError(sock->LastError());
			wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Error"));
			return FALSE;
		}
		if (sock->LastCount() > 0)
			to_write = to_write - sock->LastCount();
		else
			Sleep(500);
	}

	wxLogTrace(wxT("frTRACE_net"), wxString::Format(wxT("SendLine: Sent %s"), buf));

	wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Exit"));
	return TRUE;
}

//
// RecvLine
//
// Receives an open socket, and a pointer indicating the buffer location as
// inputs.  It reads the message from the network and stores it in a new
// buffer, updating the pointer.
//
wxUint32 FR_BackEnd::RecvLine(wxSocketBase *sock, unsigned char **buf)
{
	unsigned char   lenbuf[UINT32_LEN + 1],
	                *buffer;
	wxUint32        len,
	                to_read;

// Receive the size of the message.  On old versions it's a string
// and on this (and future) versions it's an integer.  In either case
// the value may need converting from big-endian format.
	wxLogTrace(wxT("frTRACE_functions"), wxT("RecvLine: Entry"));

	if (oldversion)
		to_read = UINT32_LEN;
	else
		to_read = sizeof(wxUint32);

	while (to_read > 0) {
		if (oldversion)
			sock->Read(&lenbuf + UINT32_LEN - to_read, to_read);
		else
			sock->Read(&len + sizeof(wxUint32) - to_read, to_read);

		if (sock->IsDisconnected()) {
			NetError(wxSOCKET_IOERR);
			wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Error"));
			return FALSE;
		}
		if (CheckExit()) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Abort"));
			return FALSE;
		}
		if ((sock->Error()) && (sock->LastError() != wxSOCKET_WOULDBLOCK)) {
			NetError(sock->LastError());
			wxLogTrace(wxT("frTRACE_functions"), wxT("RecvLine: Error"));
			return 0;
		}
		if (sock->LastCount() > 0)
			to_read = to_read - sock->LastCount();
		else
			Sleep(500);
	}

	if (oldversion) {
		memset(lenbuf + UINT32_LEN, '\0', 1);
		wxLogTrace(wxT("frTRACE_net"), wxString::Format(wxT("RecvLine: Received %s"), lenbuf));
		((wxString)wxConvCurrent->cMB2WX((const char *)lenbuf)).ToULong((unsigned long *)&len);
	} else
		wxLogTrace(wxT("frTRACE_net"), wxString::Format(wxT("RecvLine: Received %u"), len));
		
	len = wxUINT32_SWAP_ON_LE(len);
// Allocate a new buffer for the message and read it from the network.
// Since we have told the socket to wait for all data, we don't need
// to loop here.
	wxLogTrace(wxTRACE_MemAlloc, wxT("RecvLine: allocating buffer"));
	buffer = new unsigned char[len + 1];

	to_read = len;
	while (to_read > 0) {
		sock->Read(buffer + len - to_read, to_read);
		if (sock->IsDisconnected()) {
			NetError(wxSOCKET_IOERR);
			wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Error"));
			return FALSE;
		}
		if (CheckExit()) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Abort"));
			return FALSE;
		}
		if ((sock->Error()) && (sock->LastError() != wxSOCKET_WOULDBLOCK)) {
			NetError(sock->LastError());
			wxLogTrace(wxT("frTRACE_functions"), wxT("RecvLine: Error"));
			return 0;
		}
		if (sock->LastCount() > 0)
			to_read = to_read - sock->LastCount();
		else
			Sleep(500);
	}

// Append a zero termination for the buffer.  If this is a text string then
// we need the zero termination, otherwise we work from the buffer size so
// it's ignored anyway.
	memset(buffer + len, '\0', 1);
	wxLogTrace(wxT("frTRACE_net"), wxString::Format(wxT("RecvLine: Received %s"), buffer));

// Update the pointer to our new buffer.
	*buf = buffer;
	wxLogTrace(wxT("frTRACE_functions"), wxT("RecvLine: Exit"));
	return len;
}

//
// SendLineBasic
//
// Receives an open socket, a buffer and the size of the buffer as inputs.
// This send the data (without a size header and zero padded) over the
// socket.
//
bool FR_BackEnd::SendLineBasic(wxSocketBase *sock, const unsigned char *buf, const wxInt8 len)
{
	wxInt8  left,
	        to_write;

	wxLogTrace(wxT("frTRACE_functions"), wxT("SendLineBasic: Entry"));
	wxASSERT(len <= UINT32_LEN);
// Check whether the buffer needs zero padding
	left = UINT32_LEN - len;

// Send the buffer contents
	to_write = len;
	while (to_write > 0) {
		sock->Write(buf + len - to_write, to_write);
		if (sock->IsDisconnected()) {
			NetError(wxSOCKET_IOERR);
			wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Error"));
			return FALSE;
		}
		if (CheckExit()) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Abort"));
			return FALSE;
		}
		if ((sock->Error()) && (sock->LastError() != wxSOCKET_WOULDBLOCK)) {
			NetError(sock->LastError());
			wxLogTrace(wxT("frTRACE_functions"), wxT("SendLineBasic: Error"));
			return FALSE;
		}
		if (sock->LastCount() > 0)
			to_write = to_write - sock->LastCount();
		else
			Sleep(500);
	}

// If we need zero padding then send a number of zeros
	if (left > 0) {
		to_write = left;
		while (to_write > 0) {
			sock->Write("0000000000", to_write);
			if (sock->IsDisconnected()) {
				NetError(wxSOCKET_IOERR);
				wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Error"));
				return FALSE;
			}
			if (CheckExit()) {
				wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Abort"));
				return FALSE;
			}
			if ((sock->Error()) && (sock->LastError() != wxSOCKET_WOULDBLOCK)) {
				NetError(sock->LastError());
				wxLogTrace(wxT("frTRACE_functions"), wxT("SendLineBasic: Error"));
				return FALSE;
			}
			if (sock->LastCount() > 0)
				to_write = to_write - sock->LastCount();
			else
				Sleep(500);
		}
	}
	wxLogTrace(wxT("frTRACE_functions"), wxT("SendLineBasic: Exit"));
	return TRUE;
}

//
// RecvLineBasic
//
// Receives an open socket, a buffer and the size of the buffer as inputs.
// This receives the data (without a size header and zero padded) from the
// socket.
//
bool FR_BackEnd::RecvLineBasic(wxSocketBase *sock, unsigned char *buf)
{
	wxInt8 to_read;

	wxLogTrace(wxT("frTRACE_functions"), wxT("RecvLineBasic: Entry"));
// Read the data from the socket
	to_read = UINT32_LEN;
	while (to_read > 0) {
		sock->Read(buf + UINT32_LEN - to_read, to_read);
		if (sock->IsDisconnected()) {
			NetError(wxSOCKET_IOERR);
			wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Error"));
			return FALSE;
		}
		if (CheckExit()) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("SendLine: Abort"));
			return FALSE;
		}
		if ((sock->Error()) && (sock->LastError() != wxSOCKET_WOULDBLOCK)) {
			NetError(sock->LastError());
			wxLogTrace(wxT("frTRACE_functions"), wxT("RecvLineBasic: Error"));
			return FALSE;
		}
		if (sock->LastCount() > 0)
			to_read = to_read - sock->LastCount();
		else
			Sleep(500);
	}

// Zero termninate the string
	memset(buf + UINT32_LEN, '\0', 1);
	wxLogTrace(wxT("frTRACE_functions"), wxT("RecvLineBasic: Exit"));
	return TRUE;
}

//
// send_version
//
// Receives an open socket as input.  Sends the program name and version
// and verifies that the server is compatible.
//
bool FR_BackEnd::send_version(wxSocketBase *sock)
{
	unsigned char   *buffer;
	wxString        outstring,
	                rem_version;
	long            major,
	                minor;
	bool            foundnew = FALSE;

	wxLogTrace(wxT("frTRACE_functions"), wxT("send_version: Entry"));
// Send the program name and version
	outstring.Printf(wxT("%s %s"), wxT(PROGRAM), wxT(VERSION));
	if (! SendLine(sock, (const unsigned char *)(const char *)outstring.mb_str(), outstring.Length())) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("send_version: Error"));
		return FALSE;
	}
// Check whether we get an OK
	if (RecvLine(sock, &buffer) == 0) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("send_version: Error"));
		return FALSE;
	}
// If not, check whether the server is a different version of this program
	if (! wxStringEq(wxConvCurrent->cMB2WX((const char *)buffer), wxT(MSG_NET_OK))) {
		if (! wxString(wxConvCurrent->cMB2WX((const char *)buffer)).StartsWith(wxT(PROGRAM), &rem_version)) {
			SendError(wxT(ERR_INCOMP_PROG));
			wxLogTrace(wxTRACE_MemAlloc, wxT("send_version: freeing buffer"));
			delete[] buffer;
			wxLogTrace(wxT("frTRACE_functions"), wxT("send_version: Error"));
			return FALSE;
		}
		version = rem_version;
		wxLogTrace(wxTRACE_MemAlloc, wxT("send_version: freeing buffer"));
		delete[] buffer;
// If so, check whether the server uses the new or old protocol
		rem_version.Left(rem_version.Index(wxT("."))).ToLong(&major);
		rem_version = rem_version.Mid(rem_version.Index(wxT(".")) + 1);
		rem_version.Left(rem_version.Index(wxT("."))).ToLong(&minor);
		if ((major >= 1) && (minor >= 3)) {
			foundnew = TRUE;
		}
// Send an OK to say we're happy
		if (! SendLine(sock, (unsigned char *)MSG_NET_OK, Strlen(MSG_NET_OK))) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("send_version: Error"));
			return FALSE;
		}
// Check for a confirming OK from the server
		if (RecvLine(sock, &buffer) == 0) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("send_version: Error"));
			return FALSE;
		}
		if (! wxStringEq(wxConvCurrent->cMB2WX((const char *)buffer), wxT(MSG_NET_OK))) {
			wxLogTrace(wxTRACE_MemAlloc, wxT("send_version: freeing buffer"));
			delete[] buffer;
			wxLogTrace(wxT("frTRACE_functions"), wxT("send_version: Error"));
			return FALSE;
		}
	} else {
// If they're the same version then the server uses the new protocol
		foundnew = TRUE;
		version = wxT(VERSION);
	}
	wxLogTrace(wxTRACE_MemAlloc, wxT("send_version: freeing buffer"));
	delete[] buffer;
	if (foundnew) oldversion = FALSE;
	wxLogTrace(wxT("frTRACE_functions"), wxT("send_version: Exit"));
	return TRUE;
}

//
// check_version
//
// Receives an open socket as input.  Receives the program name and version
// from the socket and verifies that the client is compatible.
//
bool FR_BackEnd::check_version(wxSocketBase *sock)
{
	unsigned char   *buffer;
	wxString        outstring,
	                rem_version;
	long            major,
	                minor;
	bool            foundnew = FALSE;

	wxLogTrace(wxT("frTRACE_functions"), wxT("check_version: Entry"));
// Receive the version info from the client
	outstring.Printf(wxT("%s %s"), wxT(PROGRAM), wxT(VERSION));
	if (RecvLine(sock, &buffer) == 0) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("check_version: Error"));
		return FALSE;
	}
// Check whether it matches our version
	if (! wxStringEq(outstring, wxConvCurrent->cMB2WX((const char *)buffer))) {
// If not, check whether it's just a different version
		if (! wxString(wxConvCurrent->cMB2WX((const char *)buffer)).StartsWith(wxT(PROGRAM), &rem_version)) {
			SendError(wxT(ERR_INCOMP_PROG));
			wxLogTrace(wxTRACE_MemAlloc, wxT("check_version: freeing buffer"));
			delete[] buffer;
			wxLogTrace(wxT("frTRACE_functions"), wxT("check_version: Error"));
			return FALSE;
		}
		version = rem_version;
		wxLogTrace(wxTRACE_MemAlloc, wxT("check_version: freeing buffer"));
		delete[] buffer;
// If it is, check which protocol it uses
		rem_version.Left(rem_version.Index(wxT("."))).ToLong(&major);
		rem_version = rem_version.Mid(rem_version.Index(wxT(".")) + 1);
		rem_version.Left(rem_version.Index(wxT("."))).ToLong(&minor);
		if ((major >= 1) && (minor >= 3)) {
			foundnew = TRUE;
		}
// Send our version to the client so they can check
		if (! SendLine(sock, (const unsigned char *)(const char *)outstring.mb_str(), outstring.Length())) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("check_version: Error"));
			return FALSE;
		}
// Listen for an OK to say the client is happy
		if (RecvLine(sock, &buffer) == 0) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("check_version: Error"));
			return FALSE;
		}
		if (! wxStringEq(wxConvCurrent->cMB2WX((const char *)buffer), wxT(MSG_NET_OK))) {
			wxLogTrace(wxTRACE_MemAlloc, wxT("check_version: freeing buffer"));
			delete[] buffer;
			wxLogTrace(wxT("frTRACE_functions"), wxT("check_version: Error"));
			return FALSE;
		}
	} else {
// If the versions are the same then the client will use the new protocol
		foundnew = TRUE;
		version = wxT(VERSION);
	}
	wxLogTrace(wxTRACE_MemAlloc, wxT("check_version: freeing buffer"));
	delete[] buffer;
// Send an OK to say we're happy
	if (! SendLine(sock, (unsigned char *)MSG_NET_OK, Strlen(MSG_NET_OK))) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("check_version: Error"));
		return FALSE;
	}
	if (foundnew) oldversion = FALSE;
	wxLogTrace(wxT("frTRACE_functions"), wxT("check_version: Exit"));
	return TRUE;
}

//
// send_blocksize
//
// Receives an open socket and the blocksize to send as input.  Sends the
// blocksize info to the server.
//
bool FR_BackEnd::send_blocksize(wxSocketBase *sock, const wxUint32 blocksize)
{
	unsigned char   *buffer;
	wxString        outstring;

	wxLogTrace(wxT("frTRACE_functions"), wxT("send_blocksize: Entry"));
// Send the blocksize info
	outstring.Printf(wxT("%s%d"), wxT(MSG_NET_BLOCKSIZE), (blocksize / 1024));
	if (! SendLine(sock, (const unsigned char *)(const char *)outstring.mb_str(), outstring.Length())) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("send_blocksize: Error"));
		return FALSE;
	}
// Check for an OK from the server
	if (RecvLine(sock, &buffer) == 0) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("send_blocksize: Error"));
		return FALSE;
	}
	if (! wxStringEq(wxConvCurrent->cMB2WX((const char *)buffer), wxT(MSG_NET_OK))) {
		wxLogTrace(wxTRACE_MemAlloc, wxT("send_blocksize: freeing buffer"));
		delete[] buffer;
		wxLogTrace(wxT("frTRACE_functions"), wxT("send_blocksize: Error"));
		return FALSE;
	}
	wxLogTrace(wxTRACE_MemAlloc, wxT("send_blocksize: freeing buffer"));
	delete[] buffer;
	wxLogTrace(wxT("frTRACE_functions"), wxT("send_blocksize: Exit"));
	return TRUE;
}

//
// get_blocksize
//
// Receives an open socket as input.  Gets the blocksize info from the server
// and returns this to the calling function.
//
wxUint32 FR_BackEnd::get_blocksize(wxSocketBase *sock)
{
	unsigned char   *buffer;
	wxString        blockstring;
	unsigned long   blocksize;

	wxLogTrace(wxT("frTRACE_functions"), wxT("get_blocksize: Entry"));
// Listen for the blocksize info
	if (RecvLine(sock, &buffer) == 0) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("get_blocksize: Error"));
		return 0;
	}
	if (! ((wxString)wxConvCurrent->cMB2WX((const char *)buffer)).StartsWith(wxT(MSG_NET_BLOCKSIZE), &blockstring)) {
		SendError(wxT(ERR_PROTOCOL));
		wxLogTrace(wxTRACE_MemAlloc, wxT("get_blocksize: freeing buffer"));
		delete[] buffer;
		wxLogTrace(wxT("frTRACE_functions"), wxT("get_blocksize: Error"));
		return 0;
	}
	wxLogTrace(wxTRACE_MemAlloc, wxT("get_blocksize: freeing buffer"));
	delete[] buffer;

// Convert the blocksize to a numeric format
	blockstring.ToULong(&blocksize);
	if (blocksize == 0) {
		SendError(wxT(ERR_INVALID_BLOCK));
		wxLogTrace(wxT("frTRACE_functions"), wxT("get_blocksize: Error"));
		return 0;
	}
// Multiply it from kbytes to bytes
	blocksize = blocksize * 1024;
	if (! SendLine(sock, (unsigned char *)MSG_NET_OK, Strlen(MSG_NET_OK))) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("get_blocksize: Error"));
		return 0;
	}
	wxLogTrace(wxT("frTRACE_functions"), wxT("get_blocksize: Exit"));
	return (wxUint32)blocksize;
}

//
// send_filesize
//
// Receives an open socket and the size of the local file as input.  It sends
// this info to the server and checks whether the server file is the same
// size.
//
bool FR_BackEnd::send_filesize(wxSocketBase *sock, frFFile *infile)
{
	wxString        outstring,
	                remotesizestr;
	unsigned char   *buffer;
	frInt64         filesize,
	                remotesize;

	wxLogTrace(wxT("frTRACE_functions"), wxT("send_filesize: Entry"));
// Send the file size to the server
	filesize = infile->Length();
	outstring.Printf(wxT("%s%" wxLongLongFmtSpec "d"), wxT(MSG_NET_FILESIZE), filesize);
	if (! SendLine(sock, (const unsigned char *)(const char *)outstring.mb_str(), outstring.Length())) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("send_filesize: Error"));
		return FALSE;
	}
// Check whether the server's size matches
	if (RecvLine(sock, &buffer) == 0) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("send_filesize: Error"));
		return FALSE;
	}
	if (! wxStringEq(wxConvCurrent->cMB2WX((const char *)buffer), wxT(MSG_NET_OK))) {
// If not, the server will send an abort message with its filesize
		if (((wxString)wxConvCurrent->cMB2WX((const char *)buffer)).StartsWith(wxT(MSG_NET_ABORT), &remotesizestr)) {
			remotesize = to_int64(remotesizestr);
			wxLogTrace(wxT("frTRACE_functions"), wxT("send_filesize: Files differ"));

			wxLogTrace(wxTRACE_MemAlloc, wxT("send_filesize: freeing buffer"));
			delete[] buffer;

			if ( oldversion) {
				SendError(wxString::Format(wxT(ERR_FILESIZE1), remotesize) + wxString::Format(wxT(ERR_FILESIZE2), filesize));
				wxLogTrace(wxT("frTRACE_functions"), wxT("send_filesize: Error"));
				return FALSE;
			}
			if (! ContinueWrongFilesize(wxString::Format(wxT(ERR_FILESIZE1), remotesize) + wxString::Format(wxT(ERR_FILESIZE2), filesize))) {
				SendLine(sock, (unsigned char *)MSG_NET_ABORT, Strlen(MSG_NET_ABORT));
				wxLogTrace(wxT("frTRACE_functions"), wxT("send_filesize: Error"));
				return FALSE;
			}
			wxLogTrace(wxT("frTRACE_functions"), wxT("send_filesize: Continuing"));
			if (! infile->Resize(remotesize)) {
				SendError(wxT(ERR_RESIZE));
				wxLogTrace(wxT("frTRACE_functions"), wxT("send_filesize: Error"));
				return FALSE;
			}
			if (! SendLine(sock, (unsigned char *)MSG_NET_OK, Strlen(MSG_NET_OK))) {
				wxLogTrace(wxT("frTRACE_functions"), wxT("send_filesize: Error"));
				return FALSE;
			}
			if (RecvLine(sock, &buffer) == 0) {
				wxLogTrace(wxT("frTRACE_functions"), wxT("send_filesize: Error"));
				return FALSE;
			}
			if (! wxStringEq(wxConvCurrent->cMB2WX((const char *)buffer), wxT(MSG_NET_OK))) {
				wxLogTrace(wxT("frTRACE_functions"), wxT("send_filesize: Error"));
				wxLogTrace(wxTRACE_MemAlloc, wxT("send_filesize: freeing buffer"));
				delete[] buffer;
				return FALSE;
			}
		} else {
			SendError(wxT(ERR_PROTOCOL));
			wxLogTrace(wxTRACE_MemAlloc, wxT("send_filesize: freeing buffer"));
			delete[] buffer;
			wxLogTrace(wxT("frTRACE_functions"), wxT("send_filesize: Error"));
			return FALSE;
		}
	}
	wxLogTrace(wxTRACE_MemAlloc, wxT("send_filesize: freeing buffer"));
	delete[] buffer;
	wxLogTrace(wxT("frTRACE_functions"), wxT("send_filesize: Exit"));
	return TRUE;
}

//
// check_filesize
//
// Receives an open socket and the size of the local file as input.  It
// gets the client's filesize and checks whether they're the same.  If not
// it send the local size to the client and, if it's a new version client,
// waits to see whether the client wishes to abort, otherwise it just aborts.
//
bool FR_BackEnd::check_filesize(wxSocketBase *sock, const frInt64 localsize)
{
	wxString        outstring,
	                lenstr;
	unsigned char   *buffer;
	frInt64         remotesize;

	wxLogTrace(wxT("frTRACE_functions"), wxT("check_filesize: Entry"));
// Get the client's file size
	if (RecvLine(sock, &buffer) == 0) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("check_filesize: Error"));
		return FALSE;
	}
	if (! ((wxString)wxConvCurrent->cMB2WX((const char *)buffer)).StartsWith(wxT(MSG_NET_FILESIZE), &lenstr)) {
		SendError(wxT(ERR_PROTOCOL));
		wxLogTrace(wxTRACE_MemAlloc, wxT("check_filesize: freeing buffer"));
		delete[] buffer;
		wxLogTrace(wxT("frTRACE_functions"), wxT("check_filesize: Error"));
		return FALSE;
	}
	wxLogTrace(wxTRACE_MemAlloc, wxT("check_filesize: freeing buffer"));
	delete[] buffer;

// Check whether they're the same
	remotesize = to_int64(lenstr);
	if (localsize != remotesize) {
// If not, send our file size
		outstring.Printf(wxT("%s %" wxLongLongFmtSpec "d"), wxT(MSG_NET_ABORT), localsize);
		SendLine(sock, (const unsigned char *)(const char *)outstring.mb_str(), outstring.Length());
		wxLogTrace(wxT("frTRACE_functions"), wxT("check_filesize: Files differ"));

		if (oldversion) {
			SendError(wxString::Format(wxT(ERR_FILESIZE1), remotesize) + wxString::Format(wxT(ERR_FILESIZE2), localsize));
			wxLogTrace(wxT("frTRACE_functions"), wxT("check_filesize: Error"));
			return FALSE;
		}
		if (RecvLine(sock, &buffer) == 0) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("check_filesize: Error"));
			return FALSE;
		}
		if (! wxStringEq(wxConvCurrent->cMB2WX((const char *)buffer), wxT(MSG_NET_OK))) {
			wxLogTrace(wxTRACE_MemAlloc, wxT("check_filesize: freeing buffer"));
			delete[] buffer;
			SendError(wxString::Format(wxT(ERR_FILESIZE1), remotesize) + wxString::Format(wxT(ERR_FILESIZE2), localsize));
			wxLogTrace(wxT("frTRACE_functions"), wxT("check_filesize: Error"));
			return FALSE;
		}
		wxLogTrace(wxT("frTRACE_functions"), wxT("check_filesize: Continuing"));
	}

// Otherwise send an OK to say they match
	if (! SendLine(sock, (unsigned char *)MSG_NET_OK, Strlen(MSG_NET_OK))) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("check_filesize: Error"));
		return FALSE;
	}

	wxLogTrace(wxT("frTRACE_functions"), wxT("Check_filesize: Exit"));
	return TRUE;
}

//
// send_hashes
//
// Receives an open socket, the blocksize to use, the list of CRCs for the
// local file and the size of the local file as input.  It then sends the
// list of CRCs to the server.
//
bool FR_BackEnd::send_hashes(wxSocketBase *sock, const wxUint32 blocksize, CRCList *localcrc, const frInt64 filesize)
{
	CRCList::Node   *node;
	wxString        outstring;
	wxUint32        cnt = 0;

	wxLogTrace(wxT("frTRACE_functions"), wxT("send_hashes: Entry"));
	SendStatus(wxT(STAT_C_SEND_CRC));
// For each node in our CRC list, send the CRC to the server.
	for (node = localcrc->GetFirst(); node; node = node->GetNext()) {
		outstring.Printf(wxT("%010lu"), *(node->GetData()));
		if (! SendLineBasic(sock, (const unsigned char *)(const char *)outstring.mb_str(), 10)) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("send_hashes: Error"));
			return FALSE;
		}
		cnt++;
// Update the progress bar
		SendProgress((wxInt8)((cnt * 100) / localcrc->GetCount()));
	}
// Check whether the user has stopped the process
	if (CheckExit()) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("send_hashes: Abort"));
		return FALSE;
	}
// Send an OK to say we've finished
	if (! SendLineBasic(sock, (unsigned char *)MSG_NET_OK, Strlen(MSG_NET_OK))) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("send_hashes: Error"));
		return FALSE;
	}

	SendStatus(wxT(STAT_C_SEND_CRC_DONE));
// Reset the progress bar
	SendProgress((wxInt8)0);
	wxLogTrace(wxT("frTRACE_functions"), wxT("send_hashes: Exit"));
	return TRUE;
}

//
// get_client_crcs
//
// Receives an open socket, an empty CRClist to hold the client CRCs, the
// filesize of the local file and the blocksize we're using.  It then
// gets a list of CRCs from the client and stores them in the CRC list.
//
bool FR_BackEnd::get_client_crcs(wxSocketBase *sock, CRCList *remotecrc, const frInt64 filesize, const wxUint32 bufsize)
{
	wxUint32        cnt;
	wxLogTrace(wxTRACE_MemAlloc, wxT("get_client_crcs: allocating buffer"));
	unsigned char   *buffer = new unsigned char[UINT32_LEN + 1];
	wxString        bufferstr;
	tcrc            *crc_node;

	wxLogTrace(wxT("frTRACE_functions"), wxT("get_client_crcs: Entry"));
	SendStatus(wxT(STAT_S_REC_CRC));
	cnt = 0;
// Receive the CRCs from the client
	while (1) {
		if (! RecvLineBasic(sock, buffer)) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("get_client_crcs: Error"));
			return FALSE;
		}
		bufferstr = wxConvCurrent->cMB2WX((const char *)buffer);

// Check whether the client has finished or aborted
		if (bufferstr.StartsWith(wxT(MSG_NET_OK))) break;
		if (bufferstr.StartsWith(wxT(MSG_NET_ABORT))) {
			SendError(wxT(ERR_ABORT));
			wxLogTrace(wxTRACE_MemAlloc, wxT("get_client_crcs: freeing remotecrc"));
			remotecrc->DeleteContents(TRUE);
			wxLogTrace(wxT("frTRACE_functions"), wxT("get_client_crcs: Error"));
			return FALSE;
		}

// Add the new CRC to the end of the list
		wxLogTrace(wxTRACE_MemAlloc, wxT("get_client_crcs: allocating crc_node"));
		crc_node = new tcrc();
		bufferstr.ToULong(crc_node);
		remotecrc->Append(crc_node);
		cnt++;
// Update the progress bar
		SendProgress((wxInt8)(((frUint64)cnt * bufsize * 100) / filesize));
// Check whether the user has stopped the process
		if (CheckExit()) {
			wxLogTrace(wxTRACE_MemAlloc, wxT("get_client_crcs: freeing remotecrc"));
			remotecrc->DeleteContents(TRUE);
			wxLogTrace(wxT("frTRACE_functions"), wxT("get_client_crcs: Abort"));
			return FALSE;
		}
	}
	SendStatus(wxT(STAT_S_REC_CRC_DONE));
// Reset the progress bar
	SendProgress((wxInt8)0);
	wxLogTrace(wxTRACE_MemAlloc, wxT("get_client_crcs: freeing buffer"));
	delete[] buffer;
	wxLogTrace(wxT("frTRACE_functions"), wxT("get_client_crcs: Exit"));
	return TRUE;
}

//
// get_errcnt
//
// Receives an open socket, the local and remote CRC lists, an empty list for
// the errors and the blocksize we're using.  It finds out how many differences
// there are between the two lists, fills in the error list and returns a count
// of the differences or -1 for an error.
//
wxInt32 FR_BackEnd::get_errcnt(wxSocketBase *sock, CRCList *localcrc, CRCList *remotecrc, ErrorList *errors, const wxUint32 blocksize)
{
	wxUint32        errcnt;
	wxString        outstring;
	unsigned char   *buffer;

	wxLogTrace(wxT("frTRACE_functions"), wxT("get_errcnt: Entry"));
// Compare the CRCs
	crc_compare(localcrc, remotecrc, errors);

// Check whether the user has stopped the process
	if (CheckExit()) {
		wxLogTrace(wxTRACE_MemAlloc, wxT("get_errcnt: freeing remotecrc"));
		remotecrc->DeleteContents(TRUE);
		wxLogTrace(wxT("frTRACE_functions"), wxT("get_errcnt: Abort"));
		return -1;
	}

// Get the count
	errcnt = errors->GetCount();
	wxLogTrace(wxTRACE_MemAlloc, wxT("get_errcnt: freeing remotecrc"));
	remotecrc->DeleteContents(TRUE);

// Sends the count to the client
	outstring.Printf(wxT("%s%u"), wxT(MSG_NET_ERRORS), errcnt);
	if (! SendLine(sock, (const unsigned char *)(const char *)outstring.mb_str(), outstring.Length())) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("get_errcnt: Error"));
		return -1;
	}

// If there were no errors then we exit now
	if (errcnt == 0) {
		SendStatus(wxT(STAT_S_NO_ERRORS));
		wxLogTrace(wxT("frTRACE_functions"), wxT("get_errcnt: Exit"));
		return 0;
	}

// Otherwise, wait for an OK from the client
	if (RecvLine(sock, &buffer) == 0) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("get_errcnt: Error"));
		return -1;
	}
	if (! wxStringEq(wxConvCurrent->cMB2WX((const char *)buffer), wxT(MSG_NET_OK))) {
		wxLogTrace(wxTRACE_MemAlloc, wxT("get_errcnt: freeing buffer"));
		delete[] buffer;
		wxLogTrace(wxT("frTRACE_functions"), wxT("get_errcnt: Error"));
		return -1;
	}
	wxLogTrace(wxTRACE_MemAlloc, wxT("get_errcnt: freeing buffer"));
	delete[] buffer;
// After telling the client, report the error count to the user
	SendServerInfo(wxString::Format(wxT(STAT_S_ERR_CNT), errcnt));

	wxLogTrace(wxT("frTRACE_functions"), wxT("get_errcnt: Exit"));
	return errcnt;
}

//
// rec_errcnt
//
// Receives an open socket as input.  Gets a count of errors found from the
// server.
//
wxInt32 FR_BackEnd::rec_errcnt(wxSocketBase *sock)
{
	unsigned char   *buffer;
	wxString        errcntstr;
	unsigned long   errcnt;

	wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Entry"));
// Get the error count from the server
	if (RecvLine(sock, &buffer) == 0) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Error"));
		return -1;
	}
	if (! ((wxString)wxConvCurrent->cMB2WX((const char *)buffer)).StartsWith(wxT(MSG_NET_ERRORS), &errcntstr)) {
		SendError(wxT(ERR_PROTOCOL));
		wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Error"));
		return -1;
	}
	wxLogTrace(wxTRACE_MemAlloc, wxT("rec_errcnt: freeing buffer"));
	delete[] buffer;
// Convert the error count to a numeric value
	errcntstr.ToULong(&errcnt);
	if (errcnt == 0) {
		SendStatus(wxT(STAT_C_NO_ERRORS));
		wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Exit"));
		return 0;
	}
// Send an OK to confirm
	if (! SendLine(sock, (unsigned char *)MSG_NET_OK, Strlen(MSG_NET_OK))) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Error"));
		return -1;
	}
// Display the count to the user
	SendServerInfo(wxString::Format(wxT(STAT_C_ERR_CNT), errcnt));

	wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Exit"));
	return errcnt;
}

//
// send_segments
//
// Receives an open socket, the local CRC list, the list of errors, the
// blocksize we're using and the input file as inputs.  For each error in
// the list it sends the offset and the correct CRC, then sends the correct
// block from the local file.
//
bool FR_BackEnd::send_segments(wxSocketBase *sock, CRCList *localcrc, ErrorList *errors, const wxUint32 blocksize, frFFile *infile)
{
	wxString        outstring;
	wxUint32        cnt,
	                trycnt = 1,
			readcnt = 0,
	                errnum,
	                read_block;
	ErrorList::Node *errnode;
	unsigned char   *buffer;
	tcrc            remotecrc;

	wxLogTrace(wxT("frTRACE_functions"), wxT("send_segments: Entry"));
	SendStatus(wxT(STAT_S_SEND_DATA));
	cnt = 0;

// Start with the first error
	errnode = errors->GetFirst();
	do {
// Check whether the user has stopped the process
		if (CheckExit()) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Abort"));
			return FALSE;
		}
// Get the offset
		errnum = *(errnode->GetData());
// Send the offset and correct CRC to the client
		outstring.Printf(wxT("%s%u %010lu"), wxT(MSG_NET_BLOCK), errnum + 1, *(localcrc->Item(errnum)->GetData()));
		if (! SendLine(sock, (const unsigned char *)(const char *)outstring.mb_str(), outstring.Length())) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Error"));
			return FALSE;
		}
// Listen for an OK from the client
		if (RecvLine(sock, &buffer) == 0) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Error"));
			return FALSE;
		}
		if (! wxStringEq(wxConvCurrent->cMB2WX((const char *)buffer), wxT(MSG_NET_OK))) {
			wxLogTrace(wxTRACE_MemAlloc, wxT("send_segments: freeing buffer"));
			delete[] buffer;
			wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Error"));
			return FALSE;
		}
		wxLogTrace(wxTRACE_MemAlloc, wxT("send_segments: freeing buffer"));
		delete[] buffer;
		wxLogTrace(wxTRACE_MemAlloc, wxT("send_segments: allocating buffer"));
// Read the block from the disk
		buffer = new unsigned char[blocksize];
		do {
			read_block = get_block(buffer, blocksize, infile, errnum);
			if (read_block == 0) {
				wxLogTrace(wxTRACE_MemAlloc, wxT("send_segments: freeing buffer"));
				delete[] buffer;
				wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Error"));
				return FALSE;
			}
// Create a CRC for the block and verify it matches our earlier one
			remotecrc = createcrc(buffer, read_block, INITCRC);
			readcnt++;
		} while ((remotecrc != *(localcrc->Item(errnum)->GetData())) && (readcnt <= 10));
		if (remotecrc != *(localcrc->Item(errnum)->GetData())) {
			SendError(wxT(ERR_BLOCK_READ));
			wxLogTrace(wxTRACE_MemAlloc, wxT("send_segments: freeing buffer"));
			delete[] buffer;
			wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Error"));
			return FALSE;
		}
// Send the block to the client
		if (! SendLine(sock, buffer, read_block)) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Error"));
			return FALSE;
		}
		wxLogTrace(wxTRACE_MemAlloc, wxT("send_segments: freeing buffer"));
		delete[] buffer;
// Check for an OK from the client
		if (RecvLine(sock, &buffer) == 0) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Error"));
			return FALSE;
		}
		if (wxStringEq(wxConvCurrent->cMB2WX((const char *)buffer), wxT(MSG_NET_OK))) {
// If we get one, reset the number of attempts for this block, increment
// the number of blocks sent and move on to the next error
			errnode = errnode->GetNext();
			trycnt = 1;
			cnt++;
			SendServerInfo(wxString::Format(wxT(STAT_S_ERR_REP1), cnt, errors->GetCount()));
// Update the progress bar
			SendProgress((wxInt8)((cnt * 100) / errors->GetCount()));
// Otherwise check for a resend request from the client
		} else if (! wxStringEq(wxConvCurrent->cMB2WX((const char *)buffer), wxT(MSG_NET_RESEND))) {
// If we don't get it then abort
			SendError(wxT(ERR_INCOMP_VER));
			wxLogTrace(wxTRACE_MemAlloc, wxT("send_segments: freeing buffer"));
			delete[] buffer;
			wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Error"));
			return FALSE;
		} else {
// If we have a resend request then increment our attempts counter and loop
// back to try again
			trycnt++;
			SendServerInfo(wxString::Format(wxT(STAT_S_ERR_REP2), cnt, errors->GetCount(), trycnt));
		}
		wxLogTrace(wxTRACE_MemAlloc, wxT("send_segments: freeing buffer"));
		delete[] buffer;
	} while (errnode);
// Reset the progress bar
	SendProgress((wxInt8)0);
	wxLogTrace(wxT("frTRACE_functions"), wxT("rec_errcnt: Exit"));
	return TRUE;
}

//
// receive_blocks
//
// Receives an open socket, the blocksize we're using, the number of errors
// to fix and the input file as inputs.  Receive corrected blocks from
// the server, verify them against the CRCs sent by the server and save
// them into the file.
//
bool FR_BackEnd::receive_blocks(wxSocketBase *sock, const wxUint32 blocksize, const wxUint32 errcnt, frFFile *infile)
{
	wxUint32        cnt = 0,
	                trycnt = 1;
	wxInt32         n;
	unsigned long   crc_node,
	                write_block;
	unsigned char   *buffer;
	wxString        blocktxt,
	                temp;
	tcrc            remotecrc;


	wxLogTrace(wxT("frTRACE_functions"), wxT("receive_blocks: Entry"));
	SendStatus(wxT(STAT_C_REC_DATA));
// Keep going until we've fixed all the errors
	while (cnt < errcnt) {
// Check whether the user has stopped the process
		if (CheckExit()) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("receive_blocks: Abort"));
			return FALSE;
		}
// Get the offset and correct CRC from the server
		if (RecvLine(sock, &buffer) == 0) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("receive_blocks: Error"));
			return FALSE;
		}
		if (! ((wxString)wxConvCurrent->cMB2WX((const char *)buffer)).StartsWith(wxT(MSG_NET_BLOCK), &blocktxt)) {
			SendError(wxT(ERR_PROTOCOL));
			wxLogTrace(wxTRACE_MemAlloc, wxT("receive_blocks: freeing buffer"));
			delete[] buffer;
			wxLogTrace(wxT("frTRACE_functions"), wxT("receive_blocks: Error"));
			return FALSE;
		}
		wxLogTrace(wxTRACE_MemAlloc, wxT("receive_blocks: freeing buffer"));
		delete[] buffer;
// Convert these to numeric values
		blocktxt.BeforeFirst(wxT(' ')).ToULong(&write_block);
		write_block--;
		blocktxt.AfterFirst(wxT(' ')).ToULong(&crc_node);
// Send an OK to say we've got that info
		if (! SendLine(sock, (unsigned char *)MSG_NET_OK, Strlen(MSG_NET_OK))) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("receive_blocks: Error"));
			return FALSE;
		}
// Get the correct block from the server
		n = RecvLine(sock, &buffer);
		if (n == 0) {
			wxLogTrace(wxTRACE_MemAlloc, wxT("receive_blocks: freeing buffer"));
			delete[] buffer;
			wxLogTrace(wxT("frTRACE_functions"), wxT("receive_blocks: Error"));
			return FALSE;
		}
// Create a new CRC and verify the sent block is OK
		remotecrc = createcrc(buffer, n, INITCRC);
		if (remotecrc != crc_node) {
			wxLogTrace(wxTRACE_MemAlloc, wxT("receive_blocks: freeing buffer"));
			delete[] buffer;
// If not, request a resend and increment the attempts counter
			if (! SendLine(sock, (unsigned char *)MSG_NET_RESEND, Strlen(MSG_NET_RESEND))) {
				wxLogTrace(wxT("frTRACE_functions"), wxT("receive_blocks: Error"));
				return FALSE;
			}
			trycnt++;
			SendServerInfo(wxString::Format(wxT(STAT_C_ERR_REP2), cnt, errcnt, trycnt));
			continue;
		}
// Otherwise, save the block to the file, reset the attempts counter and
// increment the errors fixed counter.
		if (set_block(buffer, n, infile, blocksize, write_block) == 0) {
			wxLogTrace(wxTRACE_MemAlloc, wxT("receive_blocks: freeing buffer"));
			delete[] buffer;
			wxLogTrace(wxT("frTRACE_functions"), wxT("receive_blocks: Error"));
			return FALSE;
		}
		wxLogTrace(wxTRACE_MemAlloc, wxT("receive_blocks: freeing buffer"));
		delete[] buffer;;
		cnt++;
		trycnt = 1;
// Update the progress bar
		SendProgress((wxInt8)((cnt * 100) / errcnt));
		SendServerInfo(wxString::Format(wxT(STAT_C_ERR_REP1), cnt, errcnt));
// Send an OK to say we've fixed that error
		if (! SendLine(sock, (unsigned char *)MSG_NET_OK, Strlen(MSG_NET_OK))) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("receive_blocks: Error"));
			return FALSE;
		}
	}
// Reset the progress bar
	SendProgress((wxInt8)0);
	wxLogTrace(wxT("frTRACE_functions"), wxT("receive_blocks: Exit"));
	return TRUE;
}

//
// get_block
//
// Receives a buffer, the blocksize we're using, the input file and the
// offset we want to get as inputs.  It seeks to the required offset, reads in
// blocksize number of bytes (or to end-of-file) into the buffer and returns
// the number of bytes read.
//
wxUint32 FR_BackEnd::get_block(unsigned char *buffer, const wxUint32 blocksize, frFFile *infile, const wxUint32 block_num)
{
	frInt64         offset;
	wxUint32        read;

	wxLogTrace(wxT("frTRACE_functions"), wxT("get_block: Entry"));
// Calculate the offset
	offset = (frInt64)block_num * blocksize;
// Seek to this offset
	if (! infile->Seek(offset)) {
		SendError(wxString::Format(wxT(ERR_FILE_IO), wxT("Error locating block"), wxSysErrorMsg()));
		wxLogTrace(wxT("frTRACE_functions"), wxT("get_block: Error"));
		return 0;
	}
// Read blocksize bytes into the buffer
	read = infile->Read(buffer, blocksize);
// If we've not read that number of bytes, check whether we've reached the
// end of the file or had an error
	if ((read < blocksize) && (infile->Error())) {
		SendError(wxString::Format(wxT(ERR_FILE_IO), wxT("Error reading file"), wxSysErrorMsg()));
		wxLogTrace(wxT("frTRACE_functions"), wxT("get_block: Error"));
		return 0;
	}
	wxLogTrace(wxT("frTRACE_functions"), wxT("get_block: Exit"));
// Return the number of bytes read
	return read;
}

//
// set_block
//
// Receives a buffer, the blocksize we're using, the input file and the
// offset we want to get as inputs.  It seeks to the required offset and
// writes blocksize number of bytes to the file, replacing the existing
// contents.
//
wxUint32 FR_BackEnd::set_block(unsigned char *buffer, const wxUint32 size, frFFile *infile, const wxUint32 blocksize, const wxUint32 block_num)
{
	frInt64         offset;
	wxUint32        wrote;

	wxLogTrace(wxT("frTRACE_functions"), wxT("set_block: Entry"));
// Calculate the offset
	offset = (frInt64)block_num * blocksize;


// Seek to that offset
	if (! infile->Seek(offset)) {
		SendError(wxString::Format(wxT(ERR_FILE_IO), wxT("Error writing file"), wxSysErrorMsg()));
		wxLogTrace(wxT("frTRACE_functions"), wxT("set_block: Error"));
		return 0;
	}

	wxLogTrace(wxT("frTRACE_debug"), wxString::Format(wxT("Writing %d bytes to offset %" wxLongLongFmtSpec "d"), size, offset));

// Write the bytes to the file.
	if ((wrote = infile->Write(buffer, size)) < size) {
		SendError(wxString::Format(wxT(ERR_FILE_IO), wxT("Error writing file"), wxSysErrorMsg()));
		wxLogTrace(wxT("frTRACE_functions"), wxT("set_block: Error"));
		return 0;
	}
	infile->Flush();

	wxLogTrace(wxT("frTRACE_functions"), wxT("set_block: Exit"));
	return wrote;
}

//
// to_int64
//
// Receives a string.  It converts this string to an frInt64 value
// and returns this number (or 0 for a failure).
//
frInt64 FR_BackEnd::to_int64(wxString string)
{
	wxString        temp;
	unsigned long   part;
	frInt64         total = 0;
	wxUint32        cnt = 0;

	wxLogTrace(wxT("frTRACE_functions"), wxT("to_int64: Entry"));
	while (! string.Right(1).ToULong(&part)) {
		string.RemoveLast();
	}
	while (string.Len() > 0) {
		temp = string.Right(1);
		if (temp == wxT(" ")) break;
		string = string.RemoveLast();
		if (temp.ToULong(&part)) {
			total += (frInt64)(part * pow(10, cnt));
		} else {
			total = 0;
			break;
		}
		cnt++;
	}
	wxLogTrace(wxT("frTRACE_functions"), wxT("to_int64: Exit"));
	return total;
}

void FR_BackEnd::NetError(int err_code)
{
	wxString errmsg;

	switch (err_code) {
		case wxSOCKET_NOERROR:
			errmsg = wxT("Disconnected");
			break;
		case wxSOCKET_INVOP:
			errmsg = wxT("Invalid operation");
			break;
		case wxSOCKET_IOERR:
			errmsg = wxT("Input/Output error");
			break;
		case wxSOCKET_INVADDR:
			errmsg = wxT("Invalid address");
			break;
		case wxSOCKET_INVSOCK:
			errmsg = wxT("Invalid socket");
			break;
		case wxSOCKET_NOHOST:
			errmsg = wxT("No corresponding host");
			break;
		case wxSOCKET_INVPORT:
			errmsg = wxT("Invalid port");
			break;
		case wxSOCKET_WOULDBLOCK:
			errmsg = wxT("The socket is non-blocking and the operation would block");
			break;
		case wxSOCKET_TIMEDOUT:
			errmsg = wxT("The timeout for this operation expired");
			break;
		case wxSOCKET_MEMERR:
			errmsg = wxT("Memory exhausted");
			break;
	}
	SendError(wxString::Format(wxT(ERR_NETWORK), errmsg.c_str()));
}

wxString& FR_BackEnd::toBase32(const wxUint8* src, frUint64 len, wxString& dst)
{
	// Code snagged from the bitzi bitcollider
	frUint64 i,
	         index;
	wxUint8  word;
	dst.Alloc(((len * 8) / 5) + 1);

	for(i = 0, index = 0; i < len;) {
		/* Is the current word going to span a byte boundary? */
		if (index > 3) {
			word = (src[i] & (0xFF >> index));
			index = (index + 5) % 8;
			word <<= index;
			if ((i + 1) < len)
				word |= src[i + 1] >> (8 - index);

			i++;
		} else {
			word = (src[i] >> (8 - (index + 5))) & 0x1F;
			index = (index + 5) % 8;
			if (index == 0)
				i++;
		}

		wxASSERT(word < 32);
		dst += base32Alphabet[word];
	}
	return dst;
}

/* The compress function is a function. Requires smaller cache?    */
void FR_BackEnd::tiger_compress(const frUint64 *str, frUint64 state[3])
{
	tiger_compress_macro(((const frUint64*)str), ((frUint64*)state));
}

#ifdef OPTIMIZE_FOR_ALPHA
/* The compress function is inlined: works better on Alpha.        */
/* Still leaves the function above in the code, in case some other */
/* module calls it directly.                                       */
#define tiger_compress(str, state) \
  tiger_compress_macro(((frUint64*)str), ((frUint64*)state))
#endif

// Reference code for Tiger hash creation
// See http://www.cs.technion.ac.il/~biham/Reports/Tiger/
void FR_BackEnd::tiger(frUint64 *str, frUint64 length, frUint64 res[3])
{
	register frUint64 i,
	                  j;
	unsigned char     temp[64];

	res[0]=0x0123456789ABCDEFULL;
	res[1]=0xFEDCBA9876543210ULL;
	res[2]=0xF096A5B4C3B2E187ULL;

	for(i=length; i>=64; i-=64) {
#if USE_BIG_ENDIAN
		for(j=0; j<64; j++)
			temp[j^7] = ((unsigned char*)str)[j];
		tiger_compress(((frUint64*)temp), res);
#else
		tiger_compress(str, res);
#endif
		str += 8;
	}

#if USE_BIG_ENDIAN
	for(j=0; j<i; j++)
		temp[j^7] = ((unsigned char*)str)[j];

	temp[j^7] = 0x01;
	j++;
	for(; j&7; j++)
		temp[j^7] = 0;
#else
	for(j=0; j<i; j++)
		temp[j] = ((unsigned char*)str)[j];

	temp[j++] = 0x01;
	for(; j&7; j++)
		temp[j] = 0;
#endif
	if(j>56) {
		for(; j<64; j++)
			temp[j] = 0;
		tiger_compress(((frUint64*)temp), res);
		j=0;
	}
    
	for(; j<56; j++)
		temp[j] = 0;
	((frUint64*)(&(temp[56])))[0] = ((frUint64)length)<<3;
	tiger_compress(((frUint64*)temp), res);
}

// TigerTree code taken from Bitcollider
// See: http://bitzi.com/bitcollider/

/* Initialize the tigertree context */
void FR_BackEnd::tt_init(TT_CONTEXT *ctx)
{
	ctx->count = 0;
	ctx->leaf[0] = 0; // flag for leaf  calculation -- never changed
	ctx->node[0] = 1; // flag for inner node calculation -- never changed
	ctx->block = ctx->leaf + 1 ; // working area for blocks
	ctx->index = 0;   // partial block pointer/block length
	ctx->top = ctx->nodes;
}

// Combine leaf nodes
void FR_BackEnd::tt_compose(TT_CONTEXT *ctx) {
	unsigned char *node = ctx->top - NODESIZE;
	memmove((ctx->node)+1,node,NODESIZE); // copy to scratch area
	tiger((frUint64*)(ctx->node),(frUint64)(NODESIZE+1),(frUint64*)(ctx->top)); // combine two nodes
#if USE_BIG_ENDIAN
	tt_endian((unsigned char *)ctx->top);
#endif
	memmove(node,ctx->top,TIGERSIZE);           // move up result
	ctx->top -= TIGERSIZE;                      // update top ptr
}

// Process a single block
void FR_BackEnd::tt_block(TT_CONTEXT *ctx)
{
	frUint64 b;

	tiger((frUint64*)ctx->leaf,(frUint64)ctx->index+1,(frUint64*)ctx->top);
#if USE_BIG_ENDIAN
	tt_endian((unsigned char *)ctx->top);
#endif
	ctx->top += TIGERSIZE;
	++ctx->count;
	b = ctx->count;
	while(b == ((b >> 1)<<1)) { // while evenly divisible by 2...
		tt_compose(ctx);
		b = b >> 1;
	}
}

// Process the contents of the buffer
void FR_BackEnd::tt_update(TT_CONTEXT *ctx, unsigned char *buffer, wxUint32 len)
{ 

	if (ctx->index) { /* Try to fill partial block */
		unsigned left = BLOCKSIZE - ctx->index;
		if (len < left) {
			memmove(ctx->block + ctx->index, buffer, len);
			ctx->index += len;
			return; /* Finished */
		} else {
			memmove(ctx->block + ctx->index, buffer, left);
			ctx->index = BLOCKSIZE;
			tt_block(ctx);
			buffer += left;
			len -= left;
		}
	}

	while (len >= BLOCKSIZE) {
		memmove(ctx->block, buffer, BLOCKSIZE);
		ctx->index = BLOCKSIZE;
		tt_block(ctx);
		buffer += BLOCKSIZE;
		len -= BLOCKSIZE;
	}
	if ((ctx->index = len)) {   /* This assignment is intended */
        	/* Buffer leftovers */
		memmove(ctx->block, buffer, len);
	}
}

// no need to call this directly; tt_digest calls it for you
void FR_BackEnd::tt_final(TT_CONTEXT *ctx)
{
	// do last partial block, unless index is 1 (empty leaf)
	// AND we're past the first block
	if((ctx->index>0)||(ctx->top==ctx->nodes))
		tt_block(ctx);
}       

// Return the root hash
void FR_BackEnd::tt_digest(TT_CONTEXT *ctx, unsigned char *s)
{       
	tt_final(ctx);
	while( (ctx->top-TIGERSIZE) > ctx->nodes ) {
		tt_compose(ctx);
	}
	memmove(s,ctx->nodes,TIGERSIZE);
}       

// Convert to big endian format
void FR_BackEnd::tt_endian(unsigned char *s)
{
	frUint64      *i;
	unsigned char *b,
	         btemp;
	wxUint16 *w,
	         wtemp;

	for(w = (wxUint16 *)s; w < ((wxUint16 *)s) + 12; w++) {
		b = (unsigned char *)w;
		btemp = *b;
		*b = *(b + 1);
		*(b + 1) = btemp;
	}

	for(i = (frUint64 *)s; i < ((frUint64 *)s) + 3; i++) {
		w = (wxUint16 *)i;

		wtemp = *w;
		*w = *(w + 3);
		*(w + 3) = wtemp;

		wtemp = *(w + 1);
		*(w + 1) = *(w + 2);
		*(w + 2) = wtemp;
	}
}
