/*
 * (c) Copyright Robin Hill 2002
 *
 * For license terms, see the file COPYING in this directory.
 */

/* Set up typedefs for frInt64 and frUint64
 * This shouldn't be necessary but the wxWindows 64-bit types are
 * not always proper types
 */
#if wxCHECK_VERSION(2, 4, 0)
#ifdef __WINDOWS__
	typedef INT64 frInt64;
	typedef UINT64 frUint64;
#else
	typedef wxInt64 frInt64;
	typedef wxUint64 frUint64;
#endif /* __WINDOWS__ */
#else
#ifdef __WINDOWS__
	typedef INT64 frInt64;
	typedef UINT64 frUint64;
#else
	typedef long long frInt64;
	typedef unsigned long long frUint64;
#endif /* __WINDOWS__ */
#ifndef __GNUG__
	typedef wxFFile frFFile;
#endif /* __GNUG__ */
#endif /* __wxCHECK_VERSION */
/* Set up a type for the CRCs */
typedef unsigned long tcrc;
/* Declare our CRC list and Error list as linked lists */
WX_DECLARE_LIST (tcrc, CRCList);
WX_DECLARE_LIST (wxUint32, ErrorList);

/* tiger hash result size, in bytes */
#define TIGERSIZE 24

/* size of each block independently tiger-hashed, not counting leaf 0x00 prefix 
*/
#define BLOCKSIZE 1024

/* size of input to each non-leaf hash-tree node, not counting node 0x01 prefix 
*/
#define NODESIZE (TIGERSIZE*2)

/* default size of interim values stack, in TIGERSIZE
 * blocks. If this overflows (as it will for input
 * longer than 2^64 in size), havoc may ensue. */
#define STACKSIZE TIGERSIZE*56

typedef struct tt_context {
  frUint64 count;                   /* total blocks processed */
  unsigned char leaf[1+BLOCKSIZE]; /* leaf in progress */
  unsigned char *block;            /* leaf data */
  unsigned char node[1+NODESIZE]; /* node scratch space */
  int index;                      /* index into block */
  unsigned char *top;             /* top (next empty) stack slot */
  unsigned char nodes[STACKSIZE]; /* stack of interim node values */
} TT_CONTEXT;

/* This class holds all the backend functions. */
class FR_BackEnd
{
/* The public functions here are the only ones used externally. */
public:
	FR_BackEnd() {};
	virtual ~FR_BackEnd() {};
	void CreateFileHashED2k(const wxString FileName, const wxUint32 bufsize);
	void CreateFileHashMD5(const wxString FileName, const wxUint32 bufsize);
	void CreateFileHashTTH(const wxString FileName, const wxUint32 bufsize);
	void CreateFileHashCRC(const wxString FileName, const wxUint32 bufsize);
	void CreateCRC(const wxString FileName, const wxString ChkFile, const wxUint32 bufsize);
	void VerifyCRC(const wxString FileName, const wxString ChkFile);
	void Client(const wxString FileName, const wxString server, const unsigned short port, const wxUint32 bufsize, const bool passive);
	void Server(const wxString FileName, const unsigned short port, const bool passive);

/* These functions will need overriding by interface-specific functions.
 * They handle the output of information from the backend to the user.
 */
protected:
/* Error messages are sent when the application fails */
	virtual void SendError(const wxString message) = 0;
/* Alert messages are sent when the process finishes unexpectedly
 * (e.g. no errors found)
 */
	virtual void SendAlert(const wxString message) = 0;
/* Progress is an integer between 0 and 100 giving percentage completed */
	virtual void SendProgress(wxInt8 progress) = 0;
/* Status messages are general information on the current phase of the
 * process
 */
	virtual void SendStatus(const wxString message) = 0;
/* Server info messages are information messages sent by the client/server
 * process
 */
	virtual void SendServerInfo(const wxString message) = 0;
/* The hash is sent on completion of the MD5 or eDonkey processes */
	virtual void SendHash(const wxString message) = 0;
/* This checks whether the user has stopped the process.  For single-threaded
 * versions just return FALSE
 */
	virtual bool CheckExit() = 0;
/* This function checks whether to continue in the event of filesizes being
 * different.
 */
	virtual bool ContinueWrongFilesize(const wxString message) = 0;
/* This is an astracted sleep function to allow for calling different
 * sleep function when running in single or mult-threaded mode.
 */
	virtual void Sleep(unsigned long duration) = 0;

private:
	bool oldversion;
	wxString version;
	tcrc crctab[256];
	static const char base32Alphabet[];
	static const frUint64 table[];

	void init_crctab(tcrc polynomial);
	CRCList *crclist(const wxString FileName, const wxUint32 bufsize);
	tcrc createcrc(const unsigned char *buffer, const wxUint32 size, tcrc crccode);
	bool crc_compare(CRCList *correct, CRCList *check, ErrorList *errors);
	bool server_mode(wxSocketBase *sock, frFFile *infile, const wxString filename);
	bool client_mode (wxSocketBase *sock, frFFile *infile, const wxUint32 blocksize, const wxString FileName);
	bool SendLine(wxSocketBase *sock, const unsigned char *buf, const wxUint32 len);
	wxUint32 RecvLine(wxSocketBase *sock, unsigned char **buf);
	bool SendLineBasic(wxSocketBase *sock, const unsigned char *buf, const wxInt8 len);
	bool RecvLineBasic(wxSocketBase *sock, unsigned char *buf);
	bool send_version(wxSocketBase *sock);
	bool check_version(wxSocketBase *sock);
	bool send_blocksize(wxSocketBase *sock, const wxUint32 blocksize);
	wxUint32 get_blocksize(wxSocketBase *sock);
	bool send_filesize(wxSocketBase *sock, frFFile *infile);
	bool check_filesize(wxSocketBase *sock, const frInt64 localsize);
	bool send_hashes(wxSocketBase *sock, const wxUint32 blocksize, CRCList *localcrc, const frInt64 filesize);
	bool get_client_crcs(wxSocketBase *sock, CRCList *remotecrc, const frInt64 filesize, const wxUint32 bufsize);
	wxInt32 get_errcnt(wxSocketBase *sock, CRCList *localcrc, CRCList *remotecrc, ErrorList *errors, const wxUint32 blocksize);
	wxInt32 rec_errcnt(wxSocketBase *sock);
	bool send_segments(wxSocketBase *sock, CRCList *localcrc, ErrorList *errors, const wxUint32 blocksize, frFFile *infile);
	bool receive_blocks(wxSocketBase *sock, const wxUint32 blocksize, const wxUint32 errcnt, frFFile *infile);
	wxUint32 get_block(unsigned char *buffer, const wxUint32 blocksize, frFFile *infile, const wxUint32 block_num);
	wxUint32 set_block(unsigned char *buffer, const wxUint32 size, frFFile *infile, const wxUint32 blocksize, const wxUint32 block_num);
	frInt64 to_int64(wxString string);
	void NetError(int err_code);
	wxString& toBase32(const wxUint8* src, frUint64 len, wxString& dst);
	void tiger_compress(const frUint64 *str, frUint64 state[3]);
	frUint64 calcBlockSize(frUint64 FileSize, wxInt8 maxLevels);
	void tiger(frUint64 *str, frUint64 length, frUint64 res[3]);
	void tt_init(TT_CONTEXT *ctx);
	void tt_compose(TT_CONTEXT *ctx);
	void tt_block(TT_CONTEXT *ctx);
	void tt_update(TT_CONTEXT *ctx, unsigned char *buffer, wxUint32 len);
	void tt_final(TT_CONTEXT *ctx);
	void tt_digest(TT_CONTEXT *ctx, unsigned char *s);
	void tt_endian(unsigned char *s);
};
