/*
 * (c) Copyright Robin Hill 2002
 *
 * For license terms, see the file COPYING in this directory.
 */

// For compilers that support precompilation, includes "wx/wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
	#pragma hdrstop
#endif

// For all others, include the necessary headers (this file is usually all you
// need because it includes almost all "standard" wxWindows headers
#ifndef WX_PRECOMP
	#include <wx/wx.h>
#endif

#if !wxUSE_THREADS
	#error "This program requires thread support!"
#endif

#ifdef __WINDOWS__
	#include <windows.h>
#endif /* __WINDOWS */

#include <wx/notebook.h>
#include <wx/socket.h>
#include <wx/wxhtml.h>

#if wxUSE_DRAG_AND_DROP
	#include <wx/dnd.h>
#endif /* wxUSE_DRAG_AND_DROP */

#include "defines.h"
#include "frFFile.h"
#include "fr_gui.h"
#include "fr_backend.h"
#include "fr_threads_gui.h"

// Our processing thread needs a handle to the dialog and info on
// which function is being run
ProcessThread::ProcessThread(FileRepairDlg *dialog, wxUint32 function)
{
	m_dialog = dialog;
	m_function = function;
}

// Here's where the processing starts.  Read the parameters in from the
// dialog and run the relevant backend job
void *ProcessThread::Entry()
{
	wxString        FileName = m_dialog->InputFile,
	                ChkFile = m_dialog->ChkFile;
	long            port = m_dialog->port,
	                blocksize = m_dialog->blocksize;
	wxUint32        bufsize = blocksize * 1024;
	FRGUI_BackEnd   backend(m_dialog, this);

	SendThreadStart(wxT("Started"));
	switch (m_function) {
		case MODE_FH_ED2K:
			backend.CreateFileHashED2k(FileName, DEF_BLOCKSIZE);
			break;
		case MODE_FH_MD5:
			backend.CreateFileHashMD5(FileName, DEF_BLOCKSIZE);
			break;
		case MODE_FH_TTH:
			backend.CreateFileHashTTH(FileName, DEF_BLOCKSIZE);
			break;
		case MODE_FH_CRC:
			backend.CreateFileHashCRC(FileName, DEF_BLOCKSIZE);
			break;
		case MODE_CRC_CREATE:
			backend.CreateCRC(FileName, ChkFile, bufsize);
			break;
		case MODE_CRC_VERIFY:
			backend.VerifyCRC(FileName, ChkFile);
			break;
		case MODE_CLIENT_START:
			backend.Client(FileName, m_dialog->ServerAddress, (unsigned short)port, bufsize, m_dialog->passive);
			break;
		case MODE_SERVER_START:
			backend.Server(FileName, (unsigned short)port, m_dialog->passive);
			break;
	}
	SendThreadStart(wxT("Stopped"));
	m_dialog->SecondThread = NULL;
	return NULL;
}

void ProcessThread::SendThreadStart(const wxString message)
{
	wxCommandEvent event( wxEVT_COMMAND_MENU_SELECTED, THREAD_STARTED);
	event.SetInt(m_function);
	event.SetString(message);
	wxPostEvent(m_dialog, event);
}

bool ProcessThread::CheckExit()
{
	return TestDestroy();
}

FRGUI_BackEnd::FRGUI_BackEnd(FileRepairDlg *dialog, ProcessThread *thread)
{
	m_dialog = dialog;
	m_thread = thread;
}

void FRGUI_BackEnd::SendError(const wxString message)
{
	wxCommandEvent event( wxEVT_COMMAND_MENU_SELECTED, THREAD_ERROR);
	event.SetString(message);
	m_dialog->AddPendingEvent(event);
}

void FRGUI_BackEnd::SendAlert(const wxString message)
{
	wxCommandEvent event( wxEVT_COMMAND_MENU_SELECTED, THREAD_ALERT);
	event.SetString(message);
	m_dialog->AddPendingEvent(event);
}

void FRGUI_BackEnd::SendProgress(wxInt8 progress)
{
	if (progress > 100) progress = 100;
	if (progress < 0) progress = 0;
	wxCommandEvent event( wxEVT_COMMAND_MENU_SELECTED, THREAD_PROGRESS);
	event.SetInt(progress);
	m_dialog->AddPendingEvent(event);
}

void FRGUI_BackEnd::SendStatus(const wxString message)
{
	wxCommandEvent event( wxEVT_COMMAND_MENU_SELECTED, THREAD_STATUS);
	event.SetString(message);
	m_dialog->AddPendingEvent(event);
}

void FRGUI_BackEnd::SendServerInfo(const wxString message)
{
	wxCommandEvent event( wxEVT_COMMAND_MENU_SELECTED, THREAD_SRVINFO);
	event.SetString(message);
	m_dialog->AddPendingEvent(event);
}

void FRGUI_BackEnd::SendHash(const wxString message)
{
	wxCommandEvent event( wxEVT_COMMAND_MENU_SELECTED, THREAD_HASH);
	event.SetString(message);
	m_dialog->AddPendingEvent(event);
}

bool FRGUI_BackEnd::CheckExit()
{
	return m_thread->CheckExit();
}

bool FRGUI_BackEnd::ContinueWrongFilesize(const wxString message)
{
	wxString        query;
	bool            result = FALSE;

	query = message + wxT("\n\nDo you wish to continue anyway?");

	wxMutexGuiEnter();
	wxMessageDialog *CheckContinue = new wxMessageDialog(NULL, query, wxT("Continue?"), wxYES_NO | wxNO_DEFAULT | wxCENTRE | wxICON_QUESTION);

	if (CheckContinue->ShowModal() == wxID_YES)
		result = TRUE;
	delete CheckContinue;
	wxMutexGuiLeave();
	return result;
}

void FRGUI_BackEnd::Sleep(unsigned long duration)
{
	m_thread->Sleep(duration);
}
